/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
#ifdef QUANT
#include "quant.h"
#endif

double **quant = NULL;

static int setupGenDrp (int, char **, char *, int);

void  initMorgProg (char **);
void  procLodScPars (int, char **, char *);
void  chkPedFil (void);
double **setupQPed (void);
void  setPed (double **);
int   setSwitches (void);
void  setQObsData (double **);
void  allocInherit (void);

/**
 * Calls functions to do the setup steps for the Morgan programs
 * which use a Monte Carlo method.
 * This includes setting up parameters and data structures for the
 * marker and trait model,
 * reading the pedigree file and setting up the pedigree structure,
 * storing observed marker and trait data for pedigree members;
 * and setting initial values for inheritance indicators and
 * Founder Gene Labels prior to burn-in and the main MCMC iterations.
 */
int setupGenDrp (int argc, char **argv, char *ppn, int nh)
{
   int   nSwtch;

   /* initialize exit procedure, print headings.
    * fncs:  atexit(), print_head()
    * dirs:  Stuff
    */
   initMorgProg (argv);

   /* process parameter statements, set up globals resulting from these.
    * fncs:  proc_pars(), prntStatusPP(), proc_auto(), proc_fils(),
    *        proc_seeds(), prntStatusCP()
    * vars:  globals in globs.h
    * dirs:  Pars, Stuff
    */
   procLodScPars (argc, argv, ppn);

   /* check, and massage if necessary, pedigree file.
    * fncs:  chk_pedfil(), prntStatusCPF(), mak_opedfil()
    * vars:  indKeys[PedSize], compntList[nCompnt]
    * dirs:  PedComp, Stuff
    */
   chkPedFil ();

   /* read in pedigree data and set up pedigree structure.
    * fncs:  alloc_ped(), open_ped(), input_ped(), count_founders()
    * vars:  int data[PedSize][NColData], Ind nghd[PedSize+1]
    *        double quant[NTraits][PedSize], Ind nghd[].g_dat[2*(NMarks+1)]
    * dirs:  Nghds, Quant, Stuff
    */
   setPed (quant);

   nSwtch = setSwitches ();

   /* set up chromosome map for markers and trait. */
   /* fncs:  alloc_chromo(), set_chromosomes()
    * vars:  int chromo[nh+1][NMarks+1], revchro[nh+1][NMarks+1],
    *        double xecomo[2][nh+1][NMarks+1]
    * dirs:  Markers
    */
   setMaps (nh);

   /* set up data structures for simulation of inheritance and FGL.
    * fncs:  alloc_xxgg(), allocate_loci()
    * vars:  int xx0[NMarks+1][PedSize], xx1[NMarks+1][PedSize]
    *        int gg0[NMarks+1][PedSize], gg1[NMarks+1][PedSize]
    *        Locus loci[NMarks+1].*
    * dirs:  Sample
    */
   allocInherit ();

   /* initialize seeds for rans().
    * vars:  pSeedC, pSeedT
    * dirs:  Rans
    */
   setSeeds (SAMPL);

   return nSwtch;
}

/* ........................................................................ */

/**
 * Process the parameter file and set corresponding globals
 * (these start with capital letters, and are found in globs.h)
 */
void procLodScPars (int argc, char **argv, char *ppn)
{

   /* parse input parameter statements  */
   if (!proc_pars (argc, argv, ppn))
      exit (EXIT_FAILURE);

   prntStatusPP ();

   /* check statements for completeness and consistency.
    * Set up global parameters and data structures     */
   if (!proc_auto ())
      exit (EXIT_FAILURE);

   /* set up pedigree file parameters */
   if (!proc_fils ())
      exit (EXIT_FAILURE);

   /* set up seed parameters */
   if (!proc_seeds ())
      exit (EXIT_FAILURE);

   prntStatusCP ();

   return;
}

/**
 * Sets up data structures for locus and meiosis sampling.
 */
void allocInherit (void)
{

   /* allocate arrays xx0, xx1, for meiosis indicators
    *                                        int xx0 [NMarks+1][PedSize]
    *                                            xx1 [NMarks+1][PedSize]
    * allocate arrays gg0, gg1 for genes
    *                                            gg0 [NMarks+1][PedSize]
    *                                            gg1 [NMarks+1][PedSize]  */
   alloc_xxgg (NMarks, PedSize);

   if (!strcmp (ProgNam, "ibddrop"))
      return;

   /* allocate Locus structure. 
    *                                        struct Locus loci[NMarks+1]  */
   if (TraitTyp == QUANTITATIVE)
      allocate_loci (NMarks + 1, NAll, Qq, 4);
   else if (PedDatTyp == DISCR_LIAB)
      allocate_loci (NMarks + 1, NAll, Qq, 3);
   else
      allocate_loci (NMarks + 1, NAll, Qq, 1);

   return;
}
