/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/*
 AUTHOR
    written by Yun Ju Sung

 SEE ALSO
    README_twoqtl (in TwoQTL directory) for user documentation.
    Gold_twoqtl directory to see examples.

 HISTORY
    initial codes started from lm_markers.c (at 11/11/04)
    about 4000 lines of code (mostly in ../TwoQTL)   Apr  5, 2005
    extended to do componentwise                     May 20, 2005
    cvs commit done                                  Jul  5, 2005

    use two streams of random numbers,
    one for segregation indicator and
    the other for founder gene alleles               Aug  8, 2005

    evaluate the null position more
    (to get more precision)                          Aug 18, 2005

 BUGS
    (1) sampling founder gene needs testing (more?)
    (2) need to get parameter statements
    (3) proper error messages
*/

/**
 * @defgroup lm_twoqtl lm_twoqtl
 * @ingroup lodscore
 * @{
 */

#define PNAM  "lm_twoqtl"
#define TWOQTL

#include "lodscore.h"
#include "twoqtl.h"

static double *set_theta (int n_qtl);
static void finishLodSc (void);

int main (int argc, char *argv[])
{
   /* ... intermediate results      ... */

   char ***S_trait_1;                   /* paternal SI at trait positions  */
   char ***S_trait_0;                   /* maternal SI at trait positions  */

   mc_eval_t *MC_eval, *mc_eval;        /* MC evaluations for each batch   */

   int   n_qtl, n_batch, batch_size, ScorIt, null_it, debug_level, mc_size, NIt;
   int   nsw, i_max, iteration, batch, start, i, j, k;
   int  *ibase, *size_c;

   double **prob, *theta;

   /* ......  SET-UP   ................................. */

   nsw = setupAutozyg (argc, argv, PNAM, 2);
   setupSampling ();
   prob = setupInherit (1);
   set_twoqtl_parm ();                  /* TEMP */

   /* ... parameters (more later)    ... */

   n_qtl = get_twoqtl_parm_int ("n_qtl");
   theta = set_theta (n_qtl);
   n_batch = get_twoqtl_parm_int ("n_batch");
   batch_size = get_twoqtl_parm_int ("batch_size");
   ScorIt = get_twoqtl_parm_int ("ScorIt");
   null_it = get_twoqtl_parm_int ("null_it");

   debug_level = get_twoqtl_parm_int ("debug_level");

   mc_size = n_batch * batch_size;
   NIt = mc_size * ScorIt;

   ibase = get_ibase ();                /* index of first ind in each component */
   size_c = get_size_c ();              /* number of ind in each comp. */

   S_trait_1 = alloc_S_trait (NLocPos, size_c);
   S_trait_0 = alloc_S_trait (NLocPos, size_c);

   MC_eval = alloc_mc_eval (NLocPos, n_qtl);

   i_max = n_batch * n_compnts;
   if (debug_level == 4)
      for (i = 0; i < i_max; ++i)
         print_mc_eval (MC_eval[i]);

   /* ......  BURN-IN  ................................. */

   printf ("\n Now starting %d iterations of burn-in\n", NBrnIt);

   for (iteration = 0; iteration < NBrnIt; ++iteration)
   {
      if (!(iteration % ChkIt))
      {
         printf (" Starting burn-in iteration %d\n", iteration);
         fflush (stdout);
      }
      sampling (iteration, 1, NMarks, nsw, nFGL, prob);
   }

   /* ......  COMPUTE USING MCMC  ...................... */
   mc_size = 0;
   batch = 0;
   start = 1;

   for (iteration = 0; iteration < NIt; ++iteration)
   {
      sampling (iteration, 1, NMarks, nsw, nFGL, prob); /* update xx1, xx0 */
      if (debug_level == 2)
         dbgXXGG ();

      if (!((iteration + 1) % ScorIt))
      {
         j = batch * n_compnts;
         for (i = 0; i < n_compnts; ++i)
         {
            update_S_trait (i, nsw, debug_level, S_trait_1, S_trait_0);

            /* update batch sums at the alternative and null positions  */
            mc_eval = MC_eval + i + j;

            log_prob_alt (theta, 0, S_trait_1, S_trait_0, i, nsw, debug_level,
                          start, mc_eval);
            for (k = 0; k < null_it; ++k) /* more update for the null */
               log_prob_null (theta, 0, i, nsw, debug_level,
                              start && (k == 0), mc_eval);

            if (debug_level > 3)
               print_mc_eval (MC_eval[i + j]);
         }

         ++mc_size;

         /* write batch means and start new batch if current batch is done */
         if (!(mc_size % batch_size))
         {
            write_batch_means (MC_eval + batch * n_compnts, batch);
            ++batch;
            start = 1;
         }
         else
            start = 0;
      }
   }

   /* compute lod scores and write them into "lod_score" file */
   write_lod_score (MC_eval, debug_level);

   /* free */
   free_update_S_trait ();
   if (get_twoqtl_parm_dbl ("SigAdd") > 0)
      free_tq_gauss ();
   free_component ();
   free_evaluate ();

   free_S_trait (S_trait_1);
   free_S_trait (S_trait_0);

   free_mc_eval (MC_eval);
   free_twoqtl_parm ();
   free (theta);
   finishLodSc ();

   exit (EXIT_SUCCESS);
}

double *set_theta (int n_qtl)
{
   int   n_theta = 2 + n_qtl * 4;
   double *theta, *geno_means;

   theta = (double *) malloc (n_theta * sizeof (double));
   alloc_chk (WHERE, theta);

   theta[0] = get_twoqtl_parm_dbl ("SigErr");
   theta[1] = get_twoqtl_parm_dbl ("SigAdd");
   theta[2] = get_twoqtl_parm_dbl ("allele_freq_1");
   geno_means = get_geno_means ("geno_means_1");
   memcpy (theta + 3, geno_means, 3 * sizeof (double));

   if (n_qtl == 2)
   {
      theta[6] = get_twoqtl_parm_dbl ("allele_freq_2");
      geno_means = get_geno_means ("geno_means_2");
      memcpy (theta + 7, geno_means, 3 * sizeof (double));
   }
   return theta;
}

void finishLodSc (void)
{
   FILE *oSeedF;                        /* file pntr to seed file  */

   if (SamplSeedsOut)
   {
      getSeeds (SAMPL);
      if (!(oSeedF = open_fil (OSEED)))
         give_up (ERR_qit);
      writSeeds (oSeedF);
      fclose (oSeedF);
   }

   /* Free up allocated memory */
   free_xxgg ();
   free_allefrq ();
   free_chromo (0);
   free_allocate_loci ();
   free_perm ();
   locus_peel_free ();

   return;
}
