/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/**
 * This function was developed for the Morgan program "genedrop" to build
 * global data structures.  Setup for Genedrop/ programs "markerdrop"
 * and "ibddrop" is done in  proc_auto().
 * proc_drop() has as its input the results of parsing the parameter
 * statements by proc_pars().  This data is found in data structures in
 * parsepars.h and globs.h.
 * proc_drop() checks the parameters for consistency and completeness
 * and, if there are no errors, builds data structures for genedrop.
 * Functions called here are in file proc_dropf.c.  Many of these functions
 * are also called by proc_auto() and proc_lods().
 *                                                               MAJ 1997-2007
 */
int proc_drop (void)
{
   int   entry, entry2, entry3, sex;
   int   m_no, tl_no, loc_no, t_no, n;
   int   max_alle;
   int   sxsp_maps = FALSE;             /* Whether or not sex-spec maps used. */
   SIM_PTR p_s;
   PRNTMAP_PTR p_m;
   PRDIST_PTR p_l;

   if (!test_sim ())                    /* Checks for simulation or           */
      if (proc_err ())                  /* selection statement(s) present.    */
         return FALSE;

   if (n_err)
      return proc_end ();

   if (sim_traits && TRAIT_TLOC)        /* Finds the tlocs for the selected   */
      if (!find_trait_tloc ())          /* traits.                            */
         if (proc_err ())
            return FALSE;

   if (n_err)
      return proc_end ();

   if (sim_traits && TRAIT_TLOC)        /* Finds the chromosome map (with     */
      if (!find_tr_tloc_map ())         /* unlinked specs) for tlocs for      */
         if (proc_err ())               /* the requested traits.              */
            return FALSE;

   if (n_err)
      return proc_end ();

   if (sim_traits && TRAIT_TLOC)        /* Adds the lists of mapped tlocs     */
   {                                    /* to chromosomes to be simulated.    */
      if (!updt_sim_lnkd_tloc ())
         if (proc_err ())
            return FALSE;
      if (!updt_sim_unln_tloc ())       /* Adds the list of unlinked tlocs    */
         if (proc_err ())               /* to be simulated or selected.       */
            return FALSE;
   }
   if (n_err)
      return proc_end ();

   if (MARK_MAP)                        /* For each simulated chromosome,     */
   {                                    /* Checks sex-avrgd & sex-spcfc maps. */
      entry = FALSE;
      while ((p_s = nxt_sim_chrom (&entry)))
      {
         if (!test_mark_map (p_s, &sxsp_maps)) /* Checks the marker map.      */
            if (proc_err ())
               return FALSE;
         if (sim_traits && TLOC_MAP)    /* Checks the tloc map.               */
            if (!test_tloc_map (p_s, &sxsp_maps))
               if (proc_err ())
                  return FALSE;
      }
   }

   if (sim_traits && TLOC_MAP)          /* For each simulated chromosome,     */
   {                                    /* for each mapped tloc required,     */
      entry = FALSE;                    /* add to the tloc list.              */
      while ((p_s = nxt_sim_chrom (&entry)))
      {
         entry2 = FALSE;
         while ((tl_no = nxt_sim_mapd_tl_no (&entry2, p_s)))
            mak_tloc_list (tl_no, LNKD_TLOC);
      }
      entry2 = FALSE;                   /* For each unlinked tloc required,   */
      while ((tl_no = nxt_sim_unln_tl_no (&entry2))) /* add to the simulate   */
         mak_tloc_list (tl_no, UNLNKD_TLOC); /* unlinked list.                */
   }

   if (MARK_ALLELE)                     /* For each chromosome, for each      */
   {                                    /* marker, check allele frequencies.  */
      entry = FALSE;
      while ((p_s = nxt_sim_chrom (&entry)))
      {
         entry2 = FALSE;
         while ((m_no = nxt_sim_mark_no (&entry2, p_s)))
            if (!test_mark_freq (p_s, m_no))
               if (proc_err ())
                  return FALSE;
      }
   }

   if (sim_traits && TLOC_ALLELE)       /* For each tloc, check allele        */
   {                                    /* frequencies.                       */
      entry2 = FALSE;
      while ((tl_no = nxt_sim_tl_no (&entry2)))
         if (!test_tloc_freq (tl_no))
            if (proc_err ())
               return FALSE;
   }

   if (sim_traits && TRT_TLC_GENO)      /* For each trait and tloc,           */
   {
      entry2 = FALSE;
      while ((t_no = nxt_sim_tr_no (&entry2)))
      {
         entry3 = FALSE;
         while ((tl_no = nxt_tr_tl_no (&entry3, t_no)))
            if (!test_tr_tl_genotyp (t_no, tl_no)) /* Check genotype means.   */
               if (proc_err ())
                  return FALSE;
      }
   }

   if (sim_traits && TRAIT_VARI)        /* For each trait,                    */
   {
      entry2 = FALSE;
      while ((t_no = nxt_sim_tr_no (&entry2)))
      {
         if (!test_trait_vars (t_no))   /* Check additive and residual        */
            if (proc_err ())            /* variances.                         */
               return FALSE;
      }
      if (n_err)
         return proc_end ();
   }

   if (n_warn >= 1)
      fprintf (stdout, "\n");

   if (MARK_MAP)                        /* Makes the "print map" for markers  */
   {                                    /* and tlocs (may be sex-specific).   */
      entry = FALSE;
      while ((p_s = nxt_sim_chrom (&entry)))
      {
         prnt_chrom_head (p_s);

         for (sex = FEM; sex <= AVG; ++sex) /* For each gender,               */
         {
            p_m = mak_markprmap (p_s, sex); /* For the simulated markers,     */

            if (MARK_MIN_DIST)
               if (!chk_markmindist (p_m)) /* Checks distances between        */
                  if (proc_err ())      /* markers, if required.              */
                     return FALSE;

            if (sim_traits && TLOC_MAP) /* Then, for sim tlocs, if present,   */
               mak_tlocprmap (p_s, p_m, sex);

            fix_prmap_dist (p_m);       /* Find distances between adj loci.   */

            if (sim_traits && TLOC_MAP) /* Checks tloc distances.             */
               if ((sxsp_maps && (sex == FEM || sex == MAL))
                   || (!sxsp_maps && sex == AVG))
                  if (!chk_tlocprmap (p_m))
                     if (proc_err ())
                        return FALSE;
         }

         prnt_map (sxsp_maps);          /* Prints the chromosome map.         */
         if (PrintLevel >= 4)
            fprintf (stdout, "\n");

         if (MARK_ALLELE)               /* Prints marker allele frequencies.  */
            prnt_mark_freqs (p_s);

         if (sim_traits && TLOC_ALLELE)
            prnt_tloc_freqs (p_m);      /* Prints tloc allele frequencies.    */
      }
   }

   if (sim_traits && TLOC_ALLELE)       /* Prints frequencies for unmapped    */
      if (sim_unlnkd_tlocs)             /* loci.                              */
      {
         prnt_unltloc_head ();
         prnt_unltloc_freqs ();
      }

   if (sim_traits)                      /* Print trait parameters.            */
   {
      prnt_trait_head ();

      if (TRT_TLC_GENO)
         prnt_tr_tl_genotyp ();         /* Print trait/tloc parameters.       */

      if (TRAIT_VARI)
         prnt_trait_vari ();            /* Print trait variances.             */

      if (PrintLevel >= 2)
         fprintf (stdout, "\n");
   }

   if (n_err)
      return proc_end ();

   NMarks = 0;
   NLnkdTL = 0;
   entry2 = FALSE;
   while ((p_m = nxt_prntmap (&entry2, AVG)))
   {
      NMarks += p_m->no_mark_sel;
      NLnkdTL += p_m->no_tloc_sel;
   }
   if (!sim_unlnkd_tlocs)
      NUnlnTL = 0;
   else
      NUnlnTL = sim_unlnkd_tlocs->no_tloc_sel;
   NTLocs = NLnkdTL + NUnlnTL;
   NLoci = NMarks + NTLocs;

   if (!sim_traits)
      NTraits = 0;
   else
      NTraits = sim_traits->no_traits;

   if (MARK_MAP)                        /* For F and M recombination tables,  */
   {                                    /* adds an entry for each locus.      */
      alloc_recomb ();

      for (sex = FEM; sex <= MAL; ++sex)
      {
         loc_no = 0;
         entry2 = FALSE;
         while ((p_m = nxt_prntmap (&entry2, sex)))
         {
            entry3 = FALSE;
            while ((p_l = nxt_prntmap_locus (&entry3, p_m)))
            {
               mak_recomb (p_l, loc_no, sex);
               ++loc_no;
            }
         }
         entry2 = FALSE;                /* For each unlinked tloc,            */
         while ((tl_no = nxt_sim_unln_tl_no (&entry2))) /* adds entry to F/M. */
         {
            mak_no_recomb (tl_no, loc_no, sex);
            ++loc_no;
         }
      }
   }

   if (MARK_ALLELE)
   {
      loc_no = 0;                       /* For each chromosome,               */
      entry2 = FALSE;
      while ((p_m = nxt_prntmap (&entry2, AVG)))
      {                                 /* for each locus,                    */
         entry3 = FALSE;                /* stores allele frequencies.         */
         while ((p_l = nxt_prntmap_locus (&entry3, p_m)))
         {
            mak_freq (p_m, p_l, p_l->no, loc_no);
            ++loc_no;
         }
      }
   }

   if (sim_traits && TLOC_ALLELE)
   {
      entry2 = FALSE;
      while ((tl_no = nxt_sim_unln_tl_no (&entry2))) /* For unlinked loci,    */
      {
         mak_freq (NULL, NULL, tl_no, loc_no); /* stores allele frequencies.  */
         ++loc_no;
      }
   }

   if (MARK_ALLELE)
   {
      max_alle = 0;                     /* Finds maximum number of alleles.   */
      loc_no = 0;
      entry2 = FALSE;
      while ((p_m = nxt_prntmap (&entry2, AVG)))
      {
         entry3 = FALSE;
         while ((p_l = nxt_prntmap_locus (&entry3, p_m)))
         {
            if ((PAlleleFreq + loc_no)->nAlleles > max_alle)
               max_alle = (PAlleleFreq + loc_no)->nAlleles;
            ++loc_no;
         }
      }
      WidAlle = intWid (0, max_alle);
   }

   if (sim_traits)                      /* For each trait in sim trait list,  */
      /*                                   Allocate TPARS, TOUT.              */
      /*                                   Get its tloc number.               */
      /*                                   Find tloc number in recomb list.   */
      /*                                   Store linked/unlinked in TPARS.    */
      /*                                   Store chromosome number in TPARS.  */
      /*                                   Store data position in TPARS.      */
   {
      alloc_traits ();
      n = 0;
      entry2 = FALSE;
      while ((t_no = nxt_sim_tr_no (&entry2)))
      {
         entry3 = FALSE;
         tl_no = nxt_tr_tl_no (&entry3, t_no);
         loc_no = get_tl_sim_loc_no (tl_no);
         mak_trait (t_no, tl_no, loc_no, n);
         ++n;
      }
      alloc_tloc_out ();
      n = 0;
      entry2 = FALSE;
      while ((tl_no = nxt_sim_tl_no (&entry2)))
      {
         loc_no = get_tl_sim_loc_no (tl_no);
         mak_tloc_out (tl_no, loc_no, n);
         ++n;
      }
   }
   return TRUE;
}
