/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/**
 * This function builds global data structures used by the Lodscore programs.
 *
 * proc_lods() has as its input the results of parsing the parameter statements
 * for the LR_Lods programs: lm_lods and lm_schnell, and for the Lodscore
 * programs: lm_bayes, lm_markers, lm_multiple.  proc_lods() uses many of the
 * same data structures and functions as used by proc_drop() and proc_auto().
 * This function checks parameter statement data for consistency and complete-
 * ness and, if there are no errors, builds the global data structures for the
 * Morgan main programs and subroutine libraries.
 *
 * One chromosome only is selected.
 * A set of test positions is assigned to the trait locus.
 *
 * Any combination of sex-not-specified maps or sex-specific maps may be used.
 *
 *      MAJ  revs  5/2000, 1/2002, 8/2003, 6/2005, 6/2006, 10/2007.
 *      SML  revs  9/2008.
 */

int proc_lods (void)
{
   int   entry, sex;
   int   m_no, tl_no, t_no, loc_no;
   int   sxsp_maps = FALSE;             /* Whether or not sex-spec maps used. */
   SIM_PTR p_s;
   PRNTMAP_PTR p_m;
   PRDIST_PTR p_l;
   VMAP_PTR p_vm;
   PSMAP_PTR p_si;

   if (!test_lods_sel ())               /* IMPORTANT: proc_lods expects one   */
      if (proc_err ())                  /* trait with varying test positions  */
         return FALSE;                  /* for a single trait locus.          */

   if (n_err)
      return proc_end ();

   if (TRAIT_SELCTD && TRAIT_TLOC)      /* Finds the tloc for the selected    */
      if (!find_trait_tloc ())          /* trait.                             */
         if (proc_err ())
            return FALSE;

   if (n_err)
      return proc_end ();

   if (!find_tr_tloc_vmap ())           /* Finds the chromosome map for the   */
      if (proc_err ())                  /* test positions for the tloc.       */
         return FALSE;

   if (n_err)
      return proc_end ();

   entry = FALSE;
   p_s = nxt_sim_chrom (&entry);        /* One chromosome only is selected.   */

   if (!updt_sim_vtloc (p_s))           /* Adds tloc to chromosome selected.  */
      if (proc_err ())
         return FALSE;

   if (!test_mark_map (p_s, &sxsp_maps)) /* Checks the marker map.            */
      if (proc_err ())
         return FALSE;

   if (!test_tloc_vmap (p_s, &sxsp_maps)) /* Checks tloc test positions.      */
      if (proc_err ())
         return FALSE;
   if (!test_tloc_vext (p_s))
      if (proc_err ())
         return FALSE;
   if (!test_tloc_vint (p_s))
      if (proc_err ())
         return FALSE;

   entry = FALSE;
   while ((tl_no = nxt_sim_mapd_tl_no (&entry, p_s)))
      mak_tloc_list (tl_no, LNKD_TLOC); /* Adds required tloc to the list.    */

   if (MARK_ALLELE)                     /* For each marker, check allele      */
   {                                    /* frequencies.                       */
      entry = FALSE;
      while ((m_no = nxt_sim_mark_no (&entry, p_s)))
         if (!test_mark_freq (p_s, m_no))
            if (proc_err ())
               return FALSE;
   }

   if (TRAIT_SELCTD && TLOC_ALLELE)     /* For the tloc, check allele         */
   {                                    /* frequencies.                       */
      entry = FALSE;
      while ((tl_no = nxt_sim_tl_no (&entry)))
         if (!test_tloc_freq (tl_no))
            if (proc_err ())
               return FALSE;
   }

   if (TRAIT_SELCTD)                    /* Establishes what type of trait is  */
      if (!test_trait_typ ())           /* selected.                          */
         if (proc_err ())
            return FALSE;

   entry = FALSE;
   t_no = nxt_sim_tr_no (&entry);       /* Get the trait number for the tloc. */

   if (sim_traits && TRT_TLC_GENO && TraitTyp == QUANTITATIVE)
   {
      entry = FALSE;
      while ((tl_no = nxt_tr_tl_no (&entry, t_no)))
         if (!test_tr_tl_genotyp (t_no, tl_no)) /* Check genotype means.      */
            if (proc_err ())
               return FALSE;
   }

   if (sim_traits && TRAIT_VARI && TraitTyp == QUANTITATIVE)
      if (!test_trait_vars (t_no))      /* Check additive and residual        */
         if (proc_err ())               /* variances.                         */
            return FALSE;

   if (n_err)
      return proc_end ();

   if (MARK_MAP)                        /* Makes the "print map" for the sel  */
   {                                    /* markers and tloc test positions.
                                         * Assigns loci to sex-specific maps.
                                         * For maps not present, defaults to
                                         * AVG, then F or M maps.             */
      prnt_chrom_head (p_s);

      for (sex = FEM; sex <= AVG; ++sex) /* For each gender,                  */
      {
         p_m = mak_markprmap (p_s, sex); /* First, for the selected markers,  */

         if (MARK_MIN_DIST)             /* Checks distances between markers,  */
            if (!chk_markmindist (p_m)) /* if required.                       */
               if (proc_err ())
                  return FALSE;

         if (!mak_tlocvprmap (p_s, p_m, sex)) /* Checks test positions of the */
            if (proc_err ())            /* selected tloc.                     */
               return FALSE;

         shrink_prntmap (p_m);          /* Removes non selected marker loci   */
         fix_prmap_dist (p_m);          /* and finds distances between        */
      }                                 /* adjacent markers for printing.     */

      if (n_err)
         return proc_end ();

      prnt_map (sxsp_maps);             /* Prints the chromosome map.         */
      prnt_vmap (p_m->no_mark_sel, sxsp_maps); /* Prints the tloc map.        */
      if (PrintLevel >= 4)
         fprintf (stdout, "\n");

      if (MARK_ALLELE)                  /* Prints marker allele frequencies.  */
         prnt_mark_freqs (p_s);

      entry = FALSE;
      p_vm = nxt_tl_vmap (&entry, AVG, p_s);
      if (TLOC_ALLELE)                  /* Prints tloc allele frequencies.    */
         prnt_vtloc_freqs (p_vm);
   }

   if (TRAIT_SELCTD && TraitTyp == QUANTITATIVE)
   {                                    /* List traits and prints parameters. */
      prnt_trait_head ();

      if (TRT_TLC_GENO)
         prnt_tr_tl_genotyp ();
      else if (TRAIT_TLOC)
         prnt_tr_tl_list ();
      else if (!TRAIT_VARI)
         prnt_tr_list ();

      if (TRAIT_VARI)
         prnt_trait_vari ();

      if (PrintLevel >= 2)
         fprintf (stdout, "\n");
   }

   NMarks = p_m->no_mark_sel;           /* Counts loci.                       */
   NLnkdTL = p_m->no_tloc_sel;
   if (UNLTLOC_SELCTD)
      NUnlnTL = sim_unlnkd_tlocs->no_tloc_sel;
   else
      NUnlnTL = 0;

/* When Myrna first built the proc_lods.c file, she did not include any of    */
/* the modifications or references to the global NVary variable that were in  */
/* the R2 proc_auto subroutine.  However I have found references to this      */
/* variable in a few Lodscore program subroutines.  So I restored the         */
/* following if-else block from R2 proc_auto.  SML - Sep 18, 2008.            */

   if (p_vm)                            /* Check to determine if there were   */
      NVary = 1;                        /* any varying test position maps     */
   else                                 /* found.                             */
      NVary = 0;

   NTLocs = NLnkdTL + NUnlnTL;
   NLoci = NMarks + 1;                  /* Now we count the tloc, whether or
                                           not one is used.                   */

   if (!sim_traits)
      NTraits = 0;
   else
      NTraits = sim_traits->no_traits;

   entry = FALSE;
   while ((tl_no = nxt_sim_tl_no (&entry)))
      mak_tlocnam (tl_no);              /* Add this tloc name.                */

   alloc_marek ();                      /* Stores marker recombination        */
   mak_marek ();                        /* fractions in Marek[][].            */
   mak_tarek ();                        /* Stores tloc recomb fract and pos   */
   /*                                      (or .5, .5 if no linked tloc).     */

/* This next block of code was restored from R2 proc_auto.c.  Myrna had       */
/* deleted it before creating the original proc_lods.c.  However, it looks    */
/* to me that it needs to be run whenever there are varying test positions    */
/* used in a Lodscore program.  So I restored the code, unless we can         */
/* determine it is not needed.  SML - Sep 18, 2008.                           */

   if (NVary)                           /* Sets NTPos, the number of trait    */
   {                                    /* positions.                         */
      NTPos = 0;
      p_si = first_vpmap[FEM]->p_sintv;
      while (p_si)
      {
         NTPos += p_si->no_pos;
         p_si = p_si->p_next;
      }
      NTPos1 = NTPos + 1;

      mak_locpos ();                    /* Stores info about all loci.        */
      alloc_intval ();                  /* Sets positions in IntVal[]         */
      mak_intval ();                    /* and recomb fracts in RecVal[][]    */
      mak_recval ();                    /* (arrays are 2 greater than NTPos). */
   }

/* End of restored code from proc_auto.c                                      */

   entry = FALSE;                       /* For each locus, stores allele      */
   p_m = nxt_prntmap (&entry, FEM);     /* frequencies in Qq[][].             */

   if (MARK_ALLELE || TLOC_ALLELE)      /* First markers and linked tloc      */
   {
      entry = FALSE;
      while ((p_l = nxt_prntmap_locus (&entry, p_m)))
         if (p_l->locus_typ == TLOC_LOC && INHER_SELCTD)
            mak_allefrq (NULL, NULL, 0);
         else
            mak_allefrq (p_m, p_l, p_l->no);

      if (UNLTLOC_SELCTD)               /* or unlinked tloc,                  */
      {
         entry = FALSE;
         while ((tl_no = nxt_sim_unln_tl_no (&entry)))
            mak_allefrq (NULL, NULL, tl_no);
      }

/* The following 2 lines of code were also restored from R2 proc_auto.  These */
/* 2 lines do seem to be necessary.  SML - Sep 18, 2008.                      */
      else if (NVary)                   /* or test positions map used,        */
         mak_allefrq (NULL, NULL, p_vm->no);
      else if (!NTLocs)                 /* or if there is no tloc.            */
         mak_allefrq (NULL, NULL, 0);
   }

   if (MARK_NAME)
   {
      if (!test_marknam (p_s))          /* Checks marker names.               */
         if (proc_err ())
            return FALSE;

      if (!n_err)
      {
         if (mak_marknam (p_s))         /* Stores and prints marker names,    */
            prnt_marknam (p_s);         /* if provided.                       */
      }
      else
         return proc_end ();
   }

   if (MARK_DATA)
   {
      if (!test_markdat (p_s))          /* Checks marker data.                */
         if (proc_err ())
            return FALSE;

      if (!n_err)
      {
         mak_markdat (p_s);             /* Stores and prints marker data.     */
         prnt_markdat (p_s);
      }
      else
         return proc_end ();
   }

   if (LM_BAYES)
   {
      if (!test_ps_pr ())               /* Checks pseudo-priors.              */
         if (proc_err ())
            return FALSE;
   }

   if (!test_lods_pars ())              /* Check MC computational parameters. */
      if (proc_err ())
         return FALSE;

   if (TRAIT_DATA)                      /* Checks input record parameters.    */
      if (!test_ped_trtdat ())
         if (proc_err ())
            return FALSE;
   if (n_err)
      return proc_end ();

   mak_ped_dat ();

   if (TRAIT_DATA)                      /* Sets up and reports trait pedigree */
   {                                    /* data.                              */
      prnt_trt_typ ();
      prnt_ped_dat ();
   }

   if (TRAIT_PENE)                      /* Checks incomplete penetrances for  */
   {                                    /* discete trait data.                */
      if (!test_penetrnc ())
         if (proc_err ())
            return FALSE;

      prnt_penetrnc ();                 /* Prints penetrances.                */
   }

   if (n_err)
      return proc_end ();

   if (NTraits && (TraitTyp == QUANTITATIVE)) /* Stores genotype means and    */
   {                                    /* variances.                         */
      alloc_traits ();
      entry = FALSE;
      tl_no = nxt_tr_tl_no (&entry, t_no);
      loc_no = get_tl_sim_loc_no (tl_no);
      mak_trait (t_no, tl_no, loc_no, 0);
   }

   prnt_lods_pars ();                   /* Prints MC parameters.              */

   return TRUE;
}
