/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

#define PNAM "multivar"

#include "polyem.h"

/**
 * @defgroup multivar multivar
 * @ingroup polyem
 * @{
 */

int   chfact (double *a, double *b, int n, double *ldet);
void  chsolve (double *a, double *y, double *z, int n);
void  chinv (double *a, double *b, int n);

/*
 * program to test basic em equations : case of fixed + random + error
 *
 * This May 1991 version uses the full G-inverse, instead of the G for
 * observed individuals. It is adapted for a large pedigree, and takes as
 * input the components of a sparse G-inverse, computed for all
 * individuals, and also a set of eigenvalues of the G (hence G-inverse)
 * for observed individuals.
 *
 * Evgen will now instead contain the eigenvalues of the INVERSE of the
 * G for the observed individuals
 *
 * modified (especially with regard to fixed effects) for greater
 * flexibility and compatibility ; May 1992
 *
 * Modified for correct behaviour with looped pedigrees; June 1997 (SCH)
 *
 * Modified to use lower triangle for all half-stored matrices;
 *                                                         July 1997 (SCH)
 *
 * modified, Aug 1997, by MAJ, EAT and SCH for Version 2.0 revisions:
 * use of parameter file for user-friendly specification of modelling
 * parameters;
 * use of dynamic storage for all major arrays, with memory allocated just
 * prior to use, and freed as soon as no longer needed;
 *
 * modified, July 2002 by EAT:
 *     checks for missing covariates, and ignores trait values if selected
 *            covariates are missing
 *     fixed a bug in counting observed/unobserved individuals -- there was
 *            previously a problem if all individuals are observed
 *     added code to output final phenotypic values adjusted for covariates
 */

/* Access element b[i][j] from half-stored lower triangle of symmetric */
/* matrix size n Used by mmm(), chfact(), chsolve() and chinv(). */

#define BB1(b,i,j) ((i<j)?BB(b,j,i):BB(b,i,j))

void mmm (double *m1, double *m2, double *m3, double *m_out, int nt)
{
   int   i, j, k;
   double z;

   for (i = 0; i < nt * (nt + 1) / 2; i++)
      m_out[i] = 0.0;
   for (i = 0; i < nt; i++)
   {
      for (j = 0; j < nt; j++)
      {
         for (z = 0.0, k = 0; k < nt; k++)
            z += BB1 (m1, i, k) * BB1 (m2, k, j);
         for (k = i; k < nt; k++)
            BB (m_out, k, i) += z * BB1 (m3, j, k);
      }
   }
}

void mv_chk_sgfit (Term * ginv, int nt, double **b, double *ast,
                   int NTraits, int *SigModel)
{
   int   i, j, k;
   double tmp;

   k = 0;
   for (i = 0; i < NTraits; ++i)
      for (j = 0; j <= i; ++j)
      {
         if (SigModel[k])
         {
            tmp = quarz (b[i], b[j], ginv, nt);
            SigSq[k] = tmp / Nq + ast[k];
            printf (" check sig%d%d;  trace =%11.3f,  quad =%11.3f,  "
                    "net =%8.3f\n", i + 1, j + 1, ast[k] * Nq, tmp, SigSq[k]);
         }
         else
            SigSq[k] = 0.0;
         ++k;
      }
   printf ("\n");
}

void mv_chk_tufit (double **e, double *tst, int NTraits, int *TauModel)
{
   int   i, j, k;
   double tmp;

   k = 0;

   for (i = 0; i < NTraits; ++i)
      for (j = 0; j <= i; ++j)
      {
         if (TauModel[k])
         {
            tmp = innerprod (e[i], e[j], Nq);
            TauSq[k] = tmp / Nq + tst[k];
            printf (" check tau%d%d;  trace =%11.3f,  quad =%11.3f,  "
                    "net =%8.3f\n", i + 1, j + 1, tst[k] * Nq, tmp, TauSq[k]);
         }
         else
            TauSq[k] = 0.0;
         ++k;
      }
   printf ("\n\n");
}

void mv_sig_fit (Term * ginv, int nt, double **b, double *ast,
                 int NTraits, int *SigModel)
{
   int   i, j, k;

   k = 0;
   for (i = 0; i < NTraits; ++i)
      for (j = 0; j <= i; ++j)
      {
         SigSq[k] = SigModel[k] ?
            (quarz (b[i], b[j], ginv, nt)) / Nq + ast[k] : 0.0;
         ++k;
      }
}

void mv_tau_fit (double **e, double *tst, int NTraits, int *TauModel)
{
   int   i, j, k;

   k = 0;

   for (i = 0; i < NTraits; ++i)
      for (j = 0; j <= i; ++j)
      {
         TauSq[k] = TauModel[k] ?
            (innerprod (e[i], e[j], Nq)) / Nq + tst[k] : 0.0;
         ++k;
      }
}

double ss (double ev[], double *ast, double *tst, int nt)
{
   int   i, j, dd = nt * (nt + 1) / 2;
   double cc, ldet;
   double *inv1, *inv;
   double *ss;

   inv = malloc (dd * 3 * sizeof (double));
   alloc_chk (WHERE, inv);
   inv1 = inv + dd;
   ss = inv1 + dd;

   /* find the 'lamda * A + T' matrix with the first (smallest) */
   /* eigenvlaue */

   for (j = 0; j < dd; ++j)
      ss[j] = SigSq[j] / ev[1] + TauSq[j];

   if (chfact (ss, ss, nt, &ldet))
      give_up_where (WHERE, ERR_mtx, "ss");
   cc = ldet;
   chinv (ss, inv1, nt);
   for (j = 0; j < dd; j++)
      inv[j] = inv1[j];

   for (i = 2; i < Nq + 1; ++i)
   {
      /* if the evgenvalue changes, recompute the 'lamda * A +T'
       * matrix,
       * find its det, and inverse matrix */

      if (ev[i] != ev[i - 1])
      {
         for (j = 0; j < dd; ++j)
            ss[j] = SigSq[j] / ev[i] + TauSq[j];
         if (chfact (ss, ss, nt, &ldet))
            give_up_where (WHERE, ERR_mtx, "ss");
         chinv (ss, inv1, nt);
      }
      cc += ldet;
      for (j = 0; j < dd; ++j)
         inv[j] += inv1[j];
   }

   for (j = 0; j < dd; ++j)
      inv[j] = inv[j] / Nq;

   if (ast && tst)
   {
      mmm (SigSq, inv, TauSq, ast, nt);
      mmm (TauSq, inv, TauSq, tst, nt);
      for (j = 0; j < dd; ++j)
         tst[j] = TauSq[j] - tst[j];
   }

   free ((void *) inv);

   return (cc);
}

/* Compute Cholesky factorization of symmetric matrix. Inputs:  a - Lower
 * triangle of matrix n - Size of matrix
 *
 * Outputs: b - Cholesky factor of a ldet - log determinant of a
 *
 * Returns 1 if a is not positive definite, otherwise returns 0
 *
 * Note that b can point to the same location as a, otherwise a is
 * unmodified */

int chfact (double *a, double *b, int n, double *ldet)
{
   int   i, j, k, kk;
   double zz, piv;

   kk = n * (n + 1) / 2;
   *ldet = 0.0;
   for (i = n - 1; i >= 0; i--)
   {
      zz = a[--kk];
      for (k = i + 1; k < n; k++)
         zz -= BB (b, k, i) * BB (b, k, i);
      if (zz <= 0.0)
         return -1;
      b[kk] = piv = sqrt (zz);
      *ldet += log (zz);
      for (j = i - 1; j >= 0; j--)
      {
         zz = a[--kk];
         for (k = i + 1; k < n; k++)
            zz -= BB (b, k, i) * BB (b, k, j);
         b[kk] = zz / piv;
      }
   }
   return 0;
}

/* Solve A*y=z

 * Inputs:  a - Cholesky factor of A computed by chfact()
 *          y - Right hand side
 *          vector n - Size of A and y
 *
 * Outputs: z - Solution of A*y=z
 *
 * a is not modified.  The same is true for y unless y and z point
 * to the same location, which is OK
 */

void chsolve (double *a, double *y, double *z, int n)
{
   int   i, j, kk;
   double zz;

   for (i = n - 1; i >= 0; i--)
   {
      zz = y[i];
      for (j = i + 1; j < n; j++)
         zz -= BB (a, j, i) * z[j];
      z[i] = zz / BB (a, i, i);
   }
   kk = 0;
   for (i = 0; i < n; i++)
   {
      zz = z[i];
      for (j = 0; j < i; j++)
         zz -= a[kk++] * z[j];
      z[i] = zz / a[kk++];
   }
   return;
}

/*
 * Compute inverse of symmetrix pd matrix A
 *
 * Inputs:  a - Cholesky factor of A computed by chfact() n - Size of A
 *
 * Outputs: b - (half stored) inverse of A
 *
 * Note that b can point to the same location as a, otherwise a is
 * unmodified
 */

void chinv (double *a, double *b, int n)
{
   int   i, j, k, kk;
   double zz;

   kk = n * (n + 1) / 2;
   for (i = n - 1; i >= 0; i--)
   {
      zz = BB (a, i, i);
      b[--kk] = 1.0 / zz;
      for (j = i - 1; j >= 0; j--)
      {
         zz = 0.0;
         for (k = i; k > j; k--)
            zz -= BB (a, k, j) * BB (b, i, k);
         b[--kk] = zz / BB (a, j, j);
      }
   }
   kk = n * (n + 1) / 2;
   for (i = n - 1; i >= 0; i--)
   {
      for (j = i; j >= 0; j--)
      {
         zz = 0.0;
         for (k = j; k >= 0; k--)
            zz += BB (b, i, k) * BB (b, j, k);
         b[--kk] = zz;
      }
   }
}

/* Note within this routine "xx" are the residuals from any fixed effects;
 * i.e. "y" and NOT original data x
 *
 * Re-written to allow for looped pedigrees; June 1997 (SCH)
 *
 * IMPORTANT: Assumes that ginv contains the *lower left* non-zero terms
 * of the full G-inverse.  The henderson() routine has been modified to
 * do this (originally it produced the *upper right* terms).  This, of
 * course, only affects the order in which the terms are stored. Having
 * the lower left terms means that mv_itermo() can efficiently work from the
 * bottom of the pedigree upwards, which improves initial convergence if
 * data is mainly present at the bottom of the pedigree.
 */

void mv_itermo (double **xx, double **b, int NTraits, Term * ginv, int nt)
{
   int   i, j, k, k1, n, dd, l, it;
   double z, z1, *cc, *cfact, *d, *zz, *zz1, *sigch, *tauch, *siginv, *tauinv,
      **tt;

   dd = NTraits * (NTraits + 1) / 2;

   cc = malloc (sizeof (double) * (6 * dd + 3 * NTraits + NTraits * pedsize));
   alloc_chk (WHERE, cc);
   tt = malloc (sizeof (void *) * NTraits);
   alloc_chk (WHERE, tt);

   cfact = cc + dd;
   sigch = cfact + dd;
   tauch = sigch + dd;
   siginv = tauch + dd;
   tauinv = siginv + dd;
   d = tauinv + dd;
   zz = d + NTraits;
   zz1 = zz + NTraits;
   tt[0] = zz1 + NTraits;
   for (i = 1; i < NTraits; i++)
      tt[i] = tt[i - 1] + pedsize;

   /* Factor covariance matrices, and calculate inverses */
   /* (We need both the factors and the inverses later */
   if (chfact (SigSq, sigch, NTraits, &z))
      give_up_where (WHERE, ERR_mxg, "mv_itermo");
   if (chfact (TauSq, tauch, NTraits, &z))
      give_up_where (WHERE, ERR_mxe, "mv_itermo");
   chinv (sigch, siginv, NTraits);
   chinv (tauch, tauinv, NTraits);

   for (it = 0; it < ItMo; it++)
   {
      /* Deal with observations on everyone for all traits */
      for (i = 0; i < pedsize; i++)
      {
         j = qdex[i];
         if (j)
         {
            for (l = 0; l < NTraits; l++)
               zz1[l] = xx[l][j];
            chsolve (tauch, zz1, zz, NTraits);
            for (l = 0; l < NTraits; l++)
               tt[l][i] = zz[l];
         }
         else
            for (l = 0; l < NTraits; l++)
               tt[l][i] = 0.0;
      }

      for (i = 0; i < NTraits; i++)
         d[i] = 0.0;
      n = 0;
      k = nt;                           /* Start with last term of ginv */
      /* (bottom of pedigree) */
      do
      {
         /* Pick up diagonal element of ginv */
         i = ginv[k].r1;
         z = ginv[k--].value;
         for (j = 0; j < dd; j++)
            cc[j] = siginv[j] * z;
         if (qdex[i])
            for (j = 0; j < dd; j++)
               cc[j] += tauinv[j];
         chfact (cc, cfact, NTraits, &z1);
         k1 = k;

         /* Go through and deal with off-diagonal elements (j<i) */
         while (k > 0 && ginv[k].r1 == i)
         {
            j = ginv[k].r2;
            z = ginv[k--].value;
            for (l = 0; l < NTraits; l++)
               zz1[l] = b[l][j];
            chsolve (sigch, zz1, zz, NTraits);
            for (l = 0; l < NTraits; l++)
               tt[l][i] -= z * zz[l];
         }
         /* Compute zz=inv(cc)*tt[l] */
         for (l = 0; l < NTraits; l++)
            zz1[l] = tt[l][i];
         chsolve (cfact, zz1, zz, NTraits);

         /* zz now has new breeding values for individual i */
         /* Cumulate squared Differences from old breeding values */
         for (l = 0; l < NTraits; l++)
         {
            z = b[l][i] - zz[l];
            d[l] += z * z;
            b[l][i] = zz[l];
         }

         n++;

         /* Now correct tt[l] values for all individuals j, j<i, who */
         /* have off diagonal elements with i using the new breeding */
         /* values */

         k = k1;
         chsolve (sigch, zz, zz1, NTraits);
         while (k > 0 && ginv[k].r1 == i)
         {
            j = ginv[k].r2;
            z = ginv[k--].value;
            for (l = 0; l < NTraits; l++)
               tt[l][j] -= z * zz1[l];
         }
      }
      while (k > 0);

      /* Compute measure of distance between old and new breeding */
      /* value vectors...  */

      z = 0.0;
      for (l = 0; l < NTraits; l++)
         z += d[l] / SigSq[l * (l + 1) / 2 + l];
      z /= (double) (n * NTraits);

      /* ...and check if converged yet */
      if (z <= Diff)
         break;
   }
   free (tt);
   free (cc);
}

void mv_breeding_values (double **z, double **a, double **b, double **e,
                         int NTraits, Term * ginv, int nt)
{
   int   i, j;

   mv_itermo (z, b, NTraits, ginv, nt);

   for (j = 0; j < NTraits; ++j)
   {
      for (i = 0; i < pedsize; ++i)
         a[j][qdex[i]] = b[j][i];
      for (i = 1; i <= Nq; ++i)
         e[j][i] = z[j][i] - a[j][i];
   }
}

void mult_likval (int k, double log_det, double **z, double **a, double **b,
                  Term * ginv, int nt, int NTraits)
{
   double lhood;
   double *inv, z1;
   int   i, j, m;

   inv = malloc ((NTraits + 1) * NTraits / 2 * sizeof (double));
   alloc_chk (WHERE, inv);

   lhood = NTraits * Nq * log (PPI);
   lhood += log_det;

   if (chfact (TauSq, inv, NTraits, &z1))
      give_up_where (WHERE, ERR_mxe, "mult_likval");

   chinv (inv, inv, NTraits);

   m = 0;
   for (i = 0; i < NTraits; ++i)
   {
      for (j = 0; j <= i; ++j)
      {
         if (i == j)
            lhood += inv[m] * (innerprod (z[i], z[i], Nq)
                               - innerprod (a[j], a[j], Nq));
         else
            lhood += 2.0 * inv[m] * (innerprod (z[i], z[j], Nq)
                                     - innerprod (a[i], a[j], Nq));
         ++m;
      }
   }
   if (chfact (SigSq, inv, NTraits, &z1))
      give_up_where (WHERE, ERR_mxg, "mult_likval");
   chinv (inv, inv, NTraits);

   m = 0;
   for (i = 0; i < NTraits; ++i)
   {
      for (j = 0; j <= i; ++j)
      {
         if (i == j)
            lhood -= inv[m] * quarz (b[i], b[j], ginv, nt);
         else
            lhood -= 2.0 * inv[m] * quarz (b[i], b[j], ginv, nt);
         ++m;
      }
   }

   lhood = -0.5 * lhood;

   poly_print_pars (k, 5);
   printf ("\n current log-likelihood =");
   print_val (lhood);
   printf ("\n\n\n");
   fflush (stdout);

   free ((void *) inv);
}

void mv_em_iter (int k, double **a, double **b, double **epsi, double **z,
                 Term * ginv, int nt, double *evgen, int NTraits,
                 int *SigModel, int *TauModel)
{                                       /* EM iteration */
   /* iterate for expected breeding values */
   int   i, l;
   double temp;
   double log_det;
   double *ast, *tst;

   for (l = 0; l < NTraits; ++l)
      out_fixed (z[l], l);

   mv_breeding_values (z, a, b, epsi, NTraits, ginv, nt);

   ast = malloc (NTraits * (NTraits + 1) / 2 * sizeof (double));
   alloc_chk (WHERE, ast);
   tst = malloc (NTraits * (NTraits + 1) / 2 * sizeof (double));
   alloc_chk (WHERE, tst);

   log_det = ss (evgen, ast, tst, NTraits);

   /* now evaluate likelihood */
   if ((k / ItWt) * ItWt == k)
   {
      mult_likval (k, log_det, z, a, b, ginv, nt, NTraits);
      if (ChkTrc)
      {
         mv_chk_sgfit (ginv, nt, b, ast, NTraits, SigModel);
         mv_chk_tufit (epsi, tst, NTraits, TauModel);
      }
   }

   /* now implement EM equations for new variances  */
   mv_sig_fit (ginv, nt, b, ast, NTraits, SigModel);

   mv_tau_fit (epsi, tst, NTraits, TauModel);

   /* now backfit for each class of fixed effects */
   if (NFix == 0)
   {
      if (IPtEM == 1)
      {
         for (l = 0; l < NTraits; ++l)
            out_fixed (z[l], l);
         mv_breeding_values (z, a, b, epsi, NTraits, ginv, nt);
      }
      for (l = 0; l < NTraits; ++l)
         mu[l] = 0.0;
      for (l = 0; l < NTraits; ++l)
         for (i = 1; i <= Nq; ++i)
         {
            z[l][i] = (x[l][i] - a[l][i]);
            mu[l] += z[l][i];
         }
      for (l = 0; l < NTraits; ++l)
         mu[l] = mu[l] / Nq;
      for (l = 0; l < NTraits; ++l)
         Mufti[l][0][1] = mu[l];
      if (IPtEM == 1)
      {
         ss (evgen, ast, tst, NTraits);
         mv_tau_fit (epsi, tst, NTraits, TauModel);
      }
   }
   else
   {
      for (i = 1; i <= NFix; ++i)
      {
         if (IPtEM == 1)
         {
            for (l = 0; l < NTraits; ++l)
               out_fixed (z[l], l);
            mv_breeding_values (z, a, b, epsi, NTraits, ginv, nt);
         }
         for (l = 0; l < NTraits; ++l)
            out_bar_one (a[l], z[l], x[l], i, l);
         for (l = 0; l < NTraits; ++l)
         {
            temp = fixeffit (z[l], i, l);
            mu[l] = temp;
            Mufti[l][0][1] = temp;
         }
         if (IPtEM == 1)
         {
            ss (evgen, ast, tst, NTraits);
            mv_tau_fit (epsi, tst, NTraits, TauModel);
         }
      }
   }
   free ((void *) tst);
   free ((void *) ast);
}

void mv_setvals (double **z, double **b)
{
   int   dd, l, i;

   for (l = 0; l < NTraits; ++l)
   {
      mu[l] = Mufti[l][0][1];
      out_fixed (z[l], l);
      z[l][0] = 0.0;
   }

   /* set initial breeding values */

   dd = 0;
   for (l = 0; l < NTraits; ++l)
   {
      for (i = 0; i < pedsize; ++i)
         b[l][i] = (SigSq[dd] / (TauSq[dd] + SigSq[dd])) * z[l][qdex[i]];
      dd += l + 2;
   }

   for (l = 0; l < NTraits; ++l)
      x[l][0] = 0.0;
}

/* sort the vector of eigenvalues */
void sortf (double *evgen1, double *evgen, int *n, int cpnp)
{
   int   i, j, k;

   for (i = 1; i <= cpnp; ++i)
   {

      for (k = 1; k <= (*n); ++k)
      {
         if (evgen1[i] < evgen[k])
            break;
      }

      for (j = (*n) + 1; j > k; --j)
      {
         evgen[j] = evgen[j - 1];
      }

      evgen[k] = evgen1[i];
      ++(*n);
   }
}

void nogene (int NTraits, double **a, double **z, int *TauModel)
{
   int   l, i, j, m;
   double lhood, temp;
   double tauinv[MAXQT * (MAXQT + 1) / 2];

   /* use sample means as the estimates of fixed effects */
   if (NFix == 0)
   {
      for (l = 0; l < NTraits; ++l)
         mu[l] = 0.0;
      for (l = 0; l < NTraits; ++l)
         for (i = 1; i <= Nq; ++i)
         {
            z[l][i] = (x[l][i] - a[l][i]);
            mu[l] += z[l][i];
         }
      for (l = 0; l < NTraits; ++l)
         mu[l] = mu[l] / Nq;
      for (l = 0; l < NTraits; ++l)
         Mufti[l][0][1] = mu[l];
   }
   else
   {
      for (i = 1; i <= NFix; ++i)
      {
         for (l = 0; l < NTraits; ++l)
            out_bar_one (a[l], z[l], x[l], i, l);
         for (l = 0; l < NTraits; ++l)
         {
            temp = fixeffit (z[l], i, l);
            mu[l] = temp;
            Mufti[l][0][1] = temp;
         }
      }
   }
   /* take the residuals */
   for (l = 0; l < NTraits; ++l)
      out_fixed (z[l], l);

   /* use residual sample variance-covariance as estimates of residual */
   /* variance-cobvariance */

   m = 0;
   for (i = 0; i < NTraits; ++i)
      for (j = 0; j <= i; ++j)
      {
         if (TauModel[m] == 0)
            TauSq[m] = 0.0;
         else
            TauSq[m] = innerprod (z[i], z[j], Nq) / Nq;
         SigSq[m] = 0.0;
         m++;
      }

   /* compute the likelihood */

   (void) chfact (TauSq, tauinv, NTraits, &temp);
   lhood = -.5 * (NTraits * Nq * log (PPI) + Nq * temp + NTraits * Nq);

   printf ("\n estimates of environmental model\n\n");

   poly_print_pars (0, 6);
   printf ("\n environmental model log-likelihood =");
   print_val (lhood);
   printf ("\n\n\n");
   fflush (stdout);

   return;
}

void mv_alloc_vect (double ***pa, double ***pepsi, double ***pz, double ***pb)
{
   int   i;

   *pa = malloc (NTraits * sizeof (double *));
   alloc_chk (WHERE, *pa);
   *pepsi = malloc (NTraits * sizeof (double *));
   alloc_chk (WHERE, *pepsi);
   *pz = malloc (NTraits * sizeof (double *));
   alloc_chk (WHERE, *pz);
   *pb = malloc (NTraits * sizeof (double *));
   alloc_chk (WHERE, *pb);

   *(*pa) = malloc (NTraits * (Nq + 1) * sizeof (double));
   alloc_chk (WHERE, *(*pa));
   *(*pepsi) = malloc (NTraits * (Nq + 1) * sizeof (double));
   alloc_chk (WHERE, *(*pepsi));
   *(*pz) = malloc (NTraits * (Nq + 1) * sizeof (double));
   alloc_chk (WHERE, *(*pz));
   *(*pb) = malloc (NTraits * pedsize * sizeof (double));
   alloc_chk (WHERE, *(*pb));

   for (i = 1; i < NTraits; ++i)
   {
      *(*pa + i) = *(*pa + i - 1) + Nq + 1;
      *(*pepsi + i) = *(*pepsi + i - 1) + Nq + 1;
      *(*pz + i) = *(*pz + i - 1) + Nq + 1;
      *(*pb + i) = *(*pb + i - 1) + pedsize;
      **(*pz + i) = 0.0;
   }
   return;
}

void mv_free_vect (double **a, double **epsi, double **z, double **b,
                   double *evgen, Term * ginverse)
{
   free ((void *) *a);
   free ((void *) a);
   free ((void *) *epsi);
   free ((void *) epsi);
   free ((void *) *z);
   free ((void *) z);
   free ((void *) *b);
   free ((void *) b);
   a = epsi = z = b = NULL;
   free ((void *) evgen);
   evgen = NULL;
   free ((void *) ginverse);
   ginverse = NULL;

   return;
}

int main (int argc, char *argv[])
{
   int   i, l, k, cpnp, *cdex, nevgen;
   double **a, **epsi, **z, **b;
   double *evgen, *evgen1;
   double log_det;
   int   gsize;
   Ind **cpnghd;
   Term *ginverse;
   Indlist *peopl;

#if 0
   PCleanUp = clean_poly;
   atexit (clean_morg);
#endif

   setupPoly (argc, argv, PNAM);

   /* the no. of observed individuals is known;  check against limit.
    * also the no. of fixed effects levels is known;  check against no.
    * starting values */

   if (!check_pars (Nq, max_cpsize))
      exit (EXIT_FAILURE);

   /* Allocate space for eigenvalues of G-observed,  and the non-zero
    * terms of G-full-inverse
    *
    * evgen will now instead contain the eigenvalues of the INVERSE of
    * G-obs;
    * logdet, sigtrc, and tautrc have been modified accordingly */

   ginverse = malloc (pedsize * 4 * sizeof (Term));
   alloc_chk (WHERE, ginverse);

   /* find non-zero terms of G-full-inverse for all pedigree individuals */

   gsize = henderson (ginverse);
   ginverse = realloc (ginverse, (gsize + 1) * sizeof (Term));
   alloc_chk (WHERE, ginverse);

   /* For testing: print G-inverse,                           */
   /*              if eigenvalues are not to be printed, exit */

   if (ChkGnv)
   {
      printf (" G-inverse:\n\n");
      for (i = 1; i <= gsize; ++i)
      {
         printf (" %5d %5d %5d", i, ginverse[i].r1, ginverse[i].r2);
         print_val (ginverse[i].value);
         printf ("\n");
      }
      if (!ChkEig)
         exit (EXIT_SUCCESS);
      else
         printf ("\n");
   }
   /* find or read eigenvalues of g-inverse for OBSERVED individuals  */

   if (ICh == 1)
   {
      /* cdex: component index */
      cdex = malloc (pedsize * sizeof (int));
      alloc_chk (WHERE, cdex);
      /* cpnghd: component nghd */
      cpnghd = malloc (max_cpsize * sizeof (Ind *));
      alloc_chk (WHERE, cpnghd);
      /* evgen1: eigenvaluess for  component */
      evgen1 = malloc ((max_cpsize + 1) * sizeof (double));
      alloc_chk (WHERE, evgen1);
      /* evgen: eigenvalues */
      evgen = malloc ((Nq + 1) * sizeof (double));
      alloc_chk (WHERE, evgen);

      nevgen = 0;                       /* number of evgen values in evgen[] */

      for (k = 1; k < n_compnts + 1; k++)
      {
         i = 0;
         cpnp = 0;                      /* number of observed in component */
         peopl = cmplist[k].people;
         while (peopl->next != NULL)
         {
            peopl = peopl->next;
            cpnghd[i] = peopl->current;
            if (qdex[cpnghd[i]->index] != 0)
               cpnp++;
            cdex[cpnghd[i]->index] = i;
            i++;
         }

         if (cpnp > 0)
            bigcheig (evgen1, cpnp, i, cpnghd, cdex);
         else if (cpnp == 0)
            give_warn (ERR_noo, i, cpnghd[0]->self);

         sortf (evgen1, evgen, &nevgen, cpnp);
      }
      free ((void *) evgen1);
      evgen1 = NULL;
      free ((void *) cdex);
      cdex = NULL;
      free ((void *) cpnghd);
      cpnghd = NULL;
      printf (" Eigenvalues have been computed\n\n\n");
   }
   else if (ICh == 0)
   {
      evgen = peigen;
      printf (" Eigenvalues have been read from parameter file\n\n\n");
   }
   else if (ICh == -1)
   {
      if (!input_eigen (Nq))
         exit (EXIT_FAILURE);
      else
      {
         evgen = peigen;
         printf (" Eigenvalues have been read from eigenvalues " "file\n\n\n");
      }
   }

   free_cmplists ();
   if (!FinPheno)
      free_nghds ();

   /* For testing, print terms and exit  */
   if (ChkEig)
   {
      printf (" eigenvalues of G-observed:\n\n");
      for (i = 1; i <= Nq; ++i)
      {
         print_val5 (evgen[i]);
         if (i % 10 == 0)
            printf ("\n");
      }
      printf ("\n\n");
      exit (EXIT_SUCCESS);
   }
   for (i = 1; i <= Nq; ++i)
      evgen[i] = 1.0 / evgen[i];

   /* allocate space for vectors */

   mv_alloc_vect (&a, &epsi, &z, &b);

   /* set initial values */

   setfixeff ();

   mv_setvals (z, b);

   /* full or partitioned EM ? */

   if (IPtEM == 1)
      printf (" using partitioned EM\n\n");
   else
      printf (" using full EM\n\n");

   /* now ready for EM iteration  */

   for (k = 0; k <= ItEM; ++k)
   {
      mv_em_iter (k, a, b, epsi, z, ginverse, gsize, evgen, NTraits,
                  SigModel, TauModel);
   }

   for (l = 0; l < NTraits; ++l)
      out_fixed (z[l], l);
   mv_breeding_values (z, a, b, epsi, NTraits, ginverse, gsize);
   log_det = ss (evgen, 0, 0, NTraits);
   mult_likval (k, log_det, z, a, b, ginverse, gsize, NTraits);

   /* fit an environmental model if IEn = 1 */

   if (IEn == 1)
   {
      for (l = 0; l < NTraits; ++l)
         for (i = 1; i <= Nq; ++i)
            a[l][i] = 0.0;
      nogene (NTraits, a, z, TauModel);
   }

   /* output phenotypes adjusted for final covarariate effects if FinPheno=1 */
   if (FinPheno)
   {
      for (l = 0; l < NTraits; ++l)
         out_fixed (z[l], l);
      for (i = 0; i < pedsize; ++i)
      {
         printf (" %6s %6s %6s",
                 nghd[i].self, nghd[i].pa->self, nghd[i].ma->self);
         printf (" %2d", nghd[i].sex);
         k = qdex[i];
         if (k == 0)
            for (l = 0; l < 2 * NTraits; ++l)
               printf (" 999.0 ");
         else
         {
            for (l = 0; l < NTraits; ++l)
               printf (" %6.2f", x[l][k]);
            for (l = 0; l < NTraits; ++l)
               printf (" %6.2f", z[l][k]);
         }
         printf ("\n");
         fflush (stdout);
      }
      free_nghds ();
   }

   free_input_data ();
   mv_free_vect (a, epsi, z, b, evgen, ginverse);

   exit (EXIT_SUCCESS);
}

/** @} */
