/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/*
 NAME
    write_lod_score, write_batch_means, alloc_mc_eval, free_mc_eval, 
    update_batch

 SYNOPSIS
  
    #include "twoqtl.h"
    void write_lod_score (mc_eval_t *mc_eval, int debug_level);
    void write_batch_means (mc_eval_t *mc_eval, int batch);

    mc_eval_t *alloc_mc_eval (int n_pos, int n_qtl);
    void free_mc_eval (mc_eval_t *mc_eval);

    void update_batch (double value, int start, double *foo_p, double *max_p);

 DESCRIPTION

    write_lod_score computes log probability and mc standard error, using 
    the method of batch means and writes them into file "lod_score"
    (called once after MCMC is done).

    write_batch_means writes batch means into file "batch_means"
    (called when each batch is complete).

    alloc_mc_eval allocates *mc_eval and free_mc_eval frees it.

    update_batch accumulates batch sum, called once per iteration,
    and stores into two pieces (foo and max) for stable computation.

 SEE ALSO
    twoqtl.h for the definition of struct mc_eval_t

 AUTHOR
    written by Yun Ju Sung

 BUGS
    outfile names "batch_means" "lod_score" should be configurable.
                                                                           
 HISTORY      
    documented                                  Apr 21, 2005
    write batches while running MCMC            Aug 17, 2005
    change the code in var_lod                  Aug 19, 2005    
*/

#include "twoqtl.h"

const char OUT_FILE[] = "lod_score";
static FILE *out_file;

static int nCol1;                       /* file global for formatting
                                         * column headings and data columns */
static int n_batch = -1;
static int batch_size = -1;
static int mc_eval_length = -1;
static int n_loci_1 = -1;
static int n_loci_2 = -1;

static void print_lodsd (double **lod, double **var);
static void print_null (double mean_n, double var_n);
static void print_top (int lod_flag, int comp_no);

#if 0
static void print_map (void);
static void print_lod (double **lod);
static void print_sd (double **var);
#endif

static double mean_batch (mc_eval_t * mc_eval, int loci_1, int loci_2, int alt);
static double var_batch (mc_eval_t * mc_eval, int loci_1, int loci_2,
                         int alt, double mean);

void write_lod_score (mc_eval_t * mc_eval, int debug_level)
{
   int   i, j, k;
   double mean_n, mean_a, var_a, var_n;

   double **lod_comp = alloc_dbl2d (n_loci_1, n_loci_2);
   double **var_comp = alloc_dbl2d (n_loci_1, n_loci_2);

   double **lod_all = alloc_dbl2d (n_loci_1, n_loci_2);
   double **var_all = alloc_dbl2d (n_loci_1, n_loci_2);

   double *lod_c, *var_c, *lod_l, *var_l;

   if (n_batch == -1)
      n_batch = get_twoqtl_parm_int ("n_batch");

   out_file = fopen (OUT_FILE, "w");
   if (out_file == NULL)
   {
      fprintf (stderr, "Error: cannot open %s\n", OUT_FILE);
      perror ("mc_log_prob");
      exit (EXIT_FAILURE);
   }
#if 0
   print_map ();
   fprintf (out_file,
            "\n# For each component, estimate (or MC sd) of log likelihood "
            "under the null\n# comes first, followed by an array of the "
            "same under the alternative\n"
            "# (1-D array for 1 QTL and 2-D array for 2 QTL).\n");
#endif

   for (i = 0; i < n_compnts; ++i)
   {
#if 0
      fprintf (out_file,
               "\n# estimate of log (base 10) likelihood for component %d\n",
               i + 1);
      mean_n = mean_batch (mc_eval + i, 0, 0, 0);
      fprintf (out_file, "%8.4f\n", mean_n / log (10));
#endif
      mean_n = mean_batch (mc_eval + i, 0, 0, 0);
      if (n_batch > 1)
      {
         var_n = var_batch (mc_eval + i, 0, 0, 0, mean_n);
         var_n /= n_batch * exp (2 * mean_n);
      }

      for (j = 0; j < n_loci_1; ++j)
      {
         lod_c = lod_comp[j];
         var_c = var_comp[j];

         lod_l = lod_all[j];
         var_l = var_all[j];

         for (k = 0; k < n_loci_2; ++k)
         {
            mean_a = mean_batch (mc_eval + i, j, k, 1);

            lod_c[k] = mean_a;
            lod_l[k] += mean_a - mean_n;

            if (n_batch > 1)
            {
               var_a = var_batch (mc_eval + i, j, k, 1, mean_a);
               var_a /= n_batch * exp (2 * mean_a);

               var_c[k] = var_a;
               var_l[k] += var_a + var_n;
            }
         }
      }

#if 0
      print_lod (lod_comp);

      if (n_batch > 1)
      {
         fprintf (out_file,
                  "\n# MC sd of log (base 10) likelihood for component %d\n",
                  i + 1);
         fprintf (out_file, "%8.4f\n", sqrt (var_n) / log (10));
         print_sd (var_comp);
      }
#endif
      print_top (0, i);
      print_null (mean_n, var_n);
      print_lodsd (lod_comp, var_comp);
   }

#if 0
   fprintf (out_file, "\n# estimate of lod scores for entire pedigree \n");
   print_lod (lod_all);
   fprintf (out_file, "\n# MC sd of lod scores for entire pedigree \n");
   print_sd (var_all);
#endif
   print_top (1, 0);
   print_lodsd (lod_all, var_all);

   fclose (out_file);

   free_2d ((void *) &lod_comp);
   free_2d ((void *) &var_comp);

   free_2d ((void *) &lod_all);
   free_2d ((void *) &var_all);
   return;
}

/* log of mean of batch means */
double mean_batch (mc_eval_t * mc_eval, int loci_1, int loci_2, int alt)
{
   int   i;
   double foo_n, max_n, value;

   if (n_batch == -1)
      n_batch = get_twoqtl_parm_int ("n_batch");

   for (i = 0; i < n_batch; ++i)
   {
      mc_eval_t eval = mc_eval[i * n_compnts];
      if (alt == 1)
         value = eval.max_alt[loci_1][loci_2]
            + log (eval.foo_alt[loci_1][loci_2]);
      else
         value = eval.max_null + log (eval.foo_null);

      update_batch (value, i == 0, &foo_n, &max_n);
   }
   return max_n + log (foo_n / n_batch);
}

/* sample variance of batch means */
double var_batch (mc_eval_t * mc_eval, int loci_1, int loci_2,
                  int alt, double mean)
{
   int   i;
   double variance = 0;
   double value;                        /* i-th batch mean */

   if (n_batch == -1)
      n_batch = get_twoqtl_parm_int ("n_batch");

   for (i = 0; i < n_batch; ++i)
   {
      mc_eval_t eval = mc_eval[i * n_compnts];
      if (alt == 1)
         value = eval.foo_alt[loci_1][loci_2]
            * exp (eval.max_alt[loci_1][loci_2]);
      else
         value = eval.foo_null * exp (eval.max_null);

      value -= exp (mean);
      variance += value * value;
   }
   variance /= n_batch - 1;

   return variance;
}

void print_top (int lod_flag, int comp_no)
{
   if (lod_flag == 1)
      fprintf (out_file,
               "# Lod score estimates and MC sd for entire pedigree: \n\n");
   else
      fprintf (out_file,
               "# Log (base 10) likelihood estimate and MC sd "
               "for component %d:\n\n", comp_no + 1);

   nCol1 = LocNamLen > 11 ? LocNamLen : 11;

   fprintf (out_file, "# Index %9s", "TLoc1");
   if (n_loci_2 != 1)
      fprintf (out_file, "   %9s", "TLoc2");

   if (lod_flag == 1)
      fprintf (out_file, "  %11s%10s\n\n", "LodScore", "StdErr");
   else
      fprintf (out_file, "  %11s%10s\n\n", "LogLike", "StdErr");
}

void print_map (void)
{
   int   i;

   fprintf (out_file,
            "# column 1: evaluation point      \n"
            "# column 2: 0 for marker position \n"
            "# column 3: Haldane map (cM)      \n");
   for (i = 0; i < NLocPos; ++i)
      fprintf (out_file, " %2d %2d %7.2f\n",
               i + 1, LocPos[i].locus_typ, LocPos[i].pos[MAL]);
   return;
}

void print_lodsd (double **lod, double **var)
{
   int   i, j, index = 0;
   double *lod_i, *var_i;

   for (i = 0; i < n_loci_1; ++i)
   {
      lod_i = lod[i];
      var_i = var[i];
      for (j = 0; j < n_loci_2; ++j)    /* convert into log base 10  */
      {
         fprintf (out_file, " %4d %11.3f ", ++index, LocPos[i].pos[MAL]);
         if (n_loci_2 != 1)
            fprintf (out_file, "%11.3f ", LocPos[j].pos[MAL]);
         fprintf (out_file, " %11.4f%10.4f\n", lod_i[j] / log (10),
                  sqrt (var_i[j]) / log (10));
      }
   }
   fprintf (out_file, "\n\n");
   return;
}

void print_lod (double **lod)
{
   int   i, j;
   double *out;

   for (i = 0; i < n_loci_1; ++i)
   {
      out = lod[i];
      for (j = 0; j < n_loci_2; ++j)    /* convert into log base 10  */
         fprintf (out_file, "%8.4f ", out[j] / log (10));
      fprintf (out_file, "\n");
   }
   return;
}

void print_sd (double **var)
{
   int   i, j;
   double *out;

   for (i = 0; i < n_loci_1; ++i)
   {
      out = var[i];
      for (j = 0; j < n_loci_2; ++j)    /* convert into log base 10  */
         fprintf (out_file, "%8.4f ", sqrt (out[j]) / log (10));
      fprintf (out_file, "\n");
   }
   return;
}

void print_null (double mean_n, double var_n)
{
   fprintf (out_file, "    0 %11s", "unlinked");
   if (n_loci_2 != 1)
      fprintf (out_file, " %11s", "unlinked");
   fprintf (out_file, "  %11.4f%10.4f\n", mean_n / log (10),
            sqrt (var_n) / log (10));
}

/********/

void write_batch_means (mc_eval_t * mc_eval, int batch)
{
   const char BATCH_FILE[] = "batch_means";
   static FILE *batch_file;
   static int null_it = -1;
   int   n_qtl, i, j, k;
   double *foo_alt, *max_alt;

   if (null_it == -1)
      null_it = get_twoqtl_parm_int ("null_it");
   if (batch_size == -1)
      batch_size = get_twoqtl_parm_int ("batch_size");

   if (batch == 0)
   {
      batch_file = fopen (BATCH_FILE, "w");
      if (batch_file == NULL)
      {
         fprintf (stderr, "Error: cannot open %s\n", BATCH_FILE);
         perror ("write_batch_means");
         exit (EXIT_FAILURE);
      }

      if (n_batch == -1)
         n_batch = get_twoqtl_parm_int ("n_batch");
      n_qtl = get_twoqtl_parm_int ("n_qtl");

      fprintf (batch_file,
               "# %d batches, %d QTL, %d components, %d alternative loci \n\n"
               "# In each batch, likelihood estimates for components are "
               "separated by blank\n# lines. For each component, the "
               "likelihood under the null comes first,\n"
               "# followed by an array of the likelihood under the "
               "alternative (1-D array\n"
               "# for 1 QTL and 2-D array for 2 QTL).\n\n",
               n_batch, n_qtl, n_compnts, n_loci_1);
   }
   fprintf (batch_file, "# likelihood from batch %d \n", batch + 1);

   for (i = 0; i < n_compnts; ++i)
   {
      mc_eval_t *eval = mc_eval + i;
      eval->foo_null /= batch_size * null_it;
      fprintf (batch_file, "%14.9e \n", eval->foo_null * exp (eval->max_null));

      for (j = 0; j < n_loci_1; ++j)
      {
         foo_alt = eval->foo_alt[j];
         max_alt = eval->max_alt[j];
         for (k = 0; k < n_loci_2; ++k)
         {
            foo_alt[k] /= batch_size;
            fprintf (batch_file, "%14.9e ", foo_alt[k] * exp (max_alt[k]));
         }
         fprintf (batch_file, "\n");
      }
      fprintf (batch_file, "\n");
   }
   fflush (batch_file);

   if (batch == n_batch - 1)
      fclose (batch_file);
   return;
}

/********/

void update_batch (double value, int start, double *foo_p, double *max_p)
{
   double max_before, max_now;

   if (start)
   {
      *foo_p = 1.0;
      *max_p = value;
   }
   else
   {
      max_before = *max_p;
      max_now = max (max_before, value);

      *foo_p *= exp (max_before - max_now);
      *foo_p += exp (value - max_now);
      *max_p = max_now;
   }

   return;
}

static mc_eval_t setup_mc_eval (void);
mc_eval_t *alloc_mc_eval (int n_loci, int n_qtl)
{
   mc_eval_t *eval;
   int   i;

   /* set static variables */
   n_batch = get_twoqtl_parm_int ("n_batch");
   mc_eval_length = n_batch * n_compnts;
   n_loci_1 = n_loci;
   n_loci_2 = (n_qtl == 1) ? 1 : n_loci;

   eval = (mc_eval_t *) malloc (mc_eval_length * sizeof (mc_eval_t));
   alloc_chk (WHERE, eval);

   for (i = 0; i < mc_eval_length; ++i)
      eval[i] = setup_mc_eval ();

   return eval;
}

mc_eval_t setup_mc_eval (void)
{
   mc_eval_t eval;

   eval.foo_alt = alloc_dbl2d (n_loci_1, n_loci_2);
   eval.max_alt = alloc_dbl2d (n_loci_1, n_loci_2);

   eval.foo_null = 0.0;
   eval.max_null = 0.0;

   eval.n_loci_1 = n_loci_1;
   eval.n_loci_2 = n_loci_2;

   return eval;
}

void free_mc_eval (mc_eval_t * mc_eval)
{
   int   i;
   for (i = 0; i < mc_eval_length; ++i)
   {
      free_2d ((void *) &mc_eval[i].foo_alt);
      free_2d ((void *) &mc_eval[i].max_alt);
   }
   free (mc_eval);

   return;
}

/*****
----------------------------------------------------------------- 
ESTIMATE AND VARIANCE 
   for estimating log P_lambda - log P_null

   our estimate is

      log hat(P_lambda) - log hat(P_null)

   and its variance is

      (sigma_lambda)^2 / n / (P_lambda)^2 + (sigma_null)^2 / n / (P_null)^2

   where hat(P_lambda) is MCMC sample mean of 
     P_lambda (Trait | segregation indicators at i-th iteration)
   and (sigma_lambda)^2 / n is estimated by the method of batch means
   (explained below).
----------------------------------------------------------------- 
[var is corrected from Exp(Y_l / mu_l - Y_0 / mu_0)^2 / n   Aug 18, 2005]   

(sketch of proof)
    
  (1) hat(P_lambda) and hat(P_null) are independent.

  (2) By the CLT,

       sqrt(n) [hat(P_lambda) - P_lambda] --> N(0, sigma^2)

    By the delta method,

       sqrt(n) [log hat(P_lambda) - log P_lambda]
                             --> N(0, sigma^2 / P_lambda^2)
    Thus 
       var [log hat(P_lambda)] = sigma^2 / n / P_lambda^2

----------------------------------------------------------------- 
THEORY FOR ESTIMATING SD VIA THE METHOD OF BATCH MEAN
 
 sum(Y_i - bar(Y))^2 / [(n_batch - 1) * n_batch ] estimates sigma^2 / n  

where Y_i is the i-th batch mean, n_batch is the number of batches,
and n is the number of iterations of MCMC used for evaluation.
----------------------------------------------------------------- 
(sketch of proof)

  for X_1, ...., X_n
      sqrt(n) (bar(X)_n - mu) --> N(0, sigma^2)

  for each batch with batch size b goes to infinity
      sqrt(b) (bar(X)_b - mu) --> N(0, sigma^2)

      var(bar(X)_b) = sigma^2 / b

  letting Y_i = bar(X)_b_i (i-th batch mean),
     sample var of batch means [ sum(Y_i - bar(Y))^2 / (n_batch - 1) ]
     estimates sigma^2 / b.

*********************************************/
