/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/*
 NAME 
    get_ibase, get_obs_ibase, get_size_c, get_obs_size_c, get_obs_index,
    get_mbase, get_n_meiosis, get_nFGL_c, 
    free_component,
    nghd_index, comp_index,

 SYNOPSIS

    #include "twoqtl.h"
    int *get_ibase (void);
    int *get_obs_ibase (void);
    int *get_size_c (void);
    int *get_obs_size_c (void);
    int **get_obs_index (void);
    int *get_mbase (int nsw);
    int *get_n_meiosis (int nsw);
    int *get_nFGL_c (void);

    void free_component (void);

    int nghd_index (int i, int comp_no);
    int comp_index (int i, int comp_no);

 DESCRIPTION   
 
    Get functions get variables for componentwise computation
    (allocating space if necessary).

    nghd_index gives the index in nghd of i-th ind in component comp_no
    and comp_index gives the index in component of i-th ind in nghd.

    free_component frees all the static data structures used here.

 SEE ALSO
    nghd_index and comp_index below

                                      documented  YJS  07/01/2005
 */

#include "twoqtl.h"

static int *ibase = NULL;               /* index of first ind. of comp    */
static int *size_c = NULL;              /* number of ind. in comp.        */
static int *obs_ibase = NULL;           /* index of first obs ind of comp */
static int *obs_size_c = NULL;          /* number of obs. ind. of comp    */
static int **obs_index = NULL;          /* index of obs. ind. in comp     */
static int *mbase = NULL;               /* index of first meiosis of comp */
static int *n_meiosis = NULL;           /* number of meioses in comp      */
static int *nFGL_c = NULL;              /* number of FGL in comp          */

static int *construct_size_c (int *ibase, int max_base);

int  *get_ibase (void)
{
   int   j, comp_no = 0;

   if (ibase == NULL)
   {
      ibase = (int *) malloc (n_compnts * sizeof (int));
      alloc_chk (WHERE, ibase);

      ibase[comp_no++] = 0;

      for (j = 0; j < pedsize; ++j)
      {
         if (nghd[j].component != comp_no)
            ibase[comp_no++] = j;

         if (nghd[j].component < comp_no)
            give_up_where (WHERE, ERR_qit);
      }
   }
   return ibase;
}

int  *get_obs_ibase (void)
{
   int   j, comp_no = 0, index = 0;

   if (obs_ibase == NULL)
   {
      obs_ibase = (int *) malloc (n_compnts * sizeof (int));
      alloc_chk (WHERE, obs_ibase);

      obs_ibase[comp_no++] = index;

      for (j = 0; j < pedsize; ++j)
         if (nghd[j].q_dat != NULL)
         {
            if (nghd[j].component != comp_no)
               obs_ibase[comp_no++] = index;

            if (nghd[j].component < comp_no)
               give_up_where (WHERE, ERR_qit);

            ++index;
         }
   }
   return obs_ibase;
}

int **get_obs_index (void)
{
   int   i, j, index, comp_size, j_nghd;
   int  *o_index;

   if (obs_index == NULL)
   {
      obs_index = (int **) malloc (n_compnts * sizeof (int *));
      alloc_chk (WHERE, obs_index);

      if (obs_size_c == NULL)
         obs_size_c = get_obs_size_c ();
      if (size_c == NULL)
         size_c = get_size_c ();

      for (i = 0; i < n_compnts; ++i)
      {
         obs_index[i] = (int *) malloc (obs_size_c[i] * sizeof (int));
         alloc_chk (WHERE, obs_index[i]);

         o_index = obs_index[i];
         comp_size = size_c[i];

         for (j = index = 0; j < comp_size; ++j)
         {
            j_nghd = nghd_index (j, i);

            if (nghd[j_nghd].q_dat != NULL)
               o_index[index++] = j;
         }
      }
   }
   return obs_index;
}

int  *get_size_c (void)
{

   if (size_c == NULL)
   {
      if (ibase == NULL)
         ibase = get_ibase ();
      size_c = construct_size_c (ibase, pedsize);
   }
   return size_c;
}

int  *get_obs_size_c (void)
{

   if (obs_size_c == NULL)
   {
      if (obs_ibase == NULL)
         obs_ibase = get_obs_ibase ();
      obs_size_c = construct_size_c (obs_ibase, Nq);
   }
   return obs_size_c;
}

int  *construct_size_c (int *ibase, int max_base)
{
   int  *size_c;
   int   i;

   size_c = (int *) malloc (n_compnts * sizeof (int));
   alloc_chk (WHERE, size_c);

   for (i = 0; i < n_compnts; i++)
   {
      size_c[i] = (i == n_compnts - 1) ? max_base : ibase[i + 1];
      size_c[i] -= ibase[i];
   }
   return size_c;
}

int  *get_mbase (int nsw)
{
   int   comp_no = 0, i;

   if (mbase == NULL)
   {
      mbase = (int *) malloc (n_compnts * sizeof (int));
      alloc_chk (WHERE, mbase);

      mbase[comp_no++] = 1;             /* meiosis is 1-origin indexing */

      for (i = 1; i <= nsw; ++i)
      {
         if (meiosis[i].cmpnt != comp_no)
            mbase[comp_no++] = i;

         if (meiosis[i].cmpnt < comp_no)
            give_up_where (WHERE, ERR_qit);
      }

      for (i = 0; i < n_compnts; ++i)   /* compare with cmplist */
         if (mbase[i] != cmplist[i + 1].meioses[0])
            give_up_where (WHERE, ERR_qit);
   }
   return mbase;
}

int  *get_n_meiosis (int nsw)
{
   int   i;

   if (n_meiosis == NULL)
   {
      if (mbase == NULL)
         mbase = get_mbase (nsw);

      n_meiosis = construct_size_c (mbase, nsw + 1);

      for (i = 0; i < n_compnts; ++i)   /* compare with cmplist */
         if (n_meiosis[i] != cmplist[i + 1].nmeioses)
            give_up_where (WHERE, ERR_qit);
   }
   return n_meiosis;
}

int  *get_nFGL_c (void)
{
   int   j, comp_no = 0, count = 0;

   if (nFGL_c == NULL)
   {
      nFGL_c = (int *) malloc (n_compnts * sizeof (int));
      alloc_chk (WHERE, nFGL_c);

      for (j = 0; j < pedsize; ++j)
      {
         if (nghd[j].component != comp_no + 1)
         {
            nFGL_c[comp_no++] = 2 * count;
            count = 0;
         }
         if (nghd[j].pa->index == INVALID_INDEX)
            ++count;
      }
      nFGL_c[comp_no] = 2 * count;

      for (j = 0; j < n_compnts; ++j)
         if (nFGL_c[j] != (cmplist[j + 1].nfou) * 2)
            give_up_where (WHERE, ERR_qit);
   }
   return nFGL_c;
}

/*
 * map i-th ind in component comp_no to index in nghd
 * and check component of this ind is comp_no
 * note: zero-origin indexing at both ends
 */
int nghd_index (int i, int comp_no)
{
   int   k;

   if (!ibase)
      ibase = get_ibase ();

   k = (i == INVALID_INDEX) ? i : i + ibase[comp_no];

   if (nghd[k].component != comp_no + 1)
      give_up_where (WHERE, ERR_qit);

   return k;
}

/*
 * map i-th ind in nghd to index in component comp_no
 *
 * without error
 *   comp_index (nghd_index (i, comp_no), comp_no) == i
 *   nghd_index (comp_index (j, comp_no), comp_no) == j
 */
int comp_index (int i, int comp_no)
{
   int   k;

   if (!ibase)
      ibase = get_ibase ();

   k = (i == INVALID_INDEX) ? i : i - ibase[comp_no];
   return k;
}

void free_component (void)
{
   int   i;

   if (ibase != NULL)
      free (ibase);
   if (size_c != NULL)
      free (size_c);
   if (obs_ibase != NULL)
      free (obs_ibase);
   if (obs_size_c != NULL)
      free (obs_size_c);
   if (obs_index != NULL)
   {
      for (i = 0; i < n_compnts; ++i)
         free (obs_index[i]);
      free (obs_index);
   }
   if (mbase != NULL)
      free (mbase);
   if (n_meiosis != NULL)
      free (n_meiosis);
   if (nFGL_c != NULL)
      free (nFGL_c);

   return;
}
