/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/*
 NAME
    evaluate, free_evaluate

 SYNOPSIS

    #include "twoqtl.h"
    double evaluate (int **FGL_a, int **FGL_b, int comp_no);
    void free_evaluate (void);

 DESCRIPTION
    evaluate log Prob (Trait | FGL_trait) for one component
    either by enumerating all allelic allocation to FGL or by
    simulating such allocation.

    free_evaluate frees q_trait (obs. quantitative trait) and geno_means_joint.

 INPUT

    FGL_1   observed FGL (founder genome label) at the first locus
    FGL_2   observed FGL at the second locus, NULL if number of QTL is 1
    comp_no 0, 1, ..., n_compnt - 1

    theta, new are reserved for future use, not used right now.

 DETAILS

    evaluate () sets

      int fou_sum_MC = get_twoqtl_parm_int ("fou_sum_MC");
      int n_fou_sum_MC = get_twoqtl_parm_int ("n_fou_sum_MC");

    if fou_sum_MC == 0, it does exact sum over founder genes and
    if fou_sum_MC == 1, it does MC sum with MC length n_fou_sum_MC.

    enumerate () goes through all possible alleles of founder genes [using
    next_allele ()], thus is exact but very slow especially for 2 QTL.
    For example, if number of founders are 10, then number of possible
    alleles are 2^20 (== 1,048,576) for each QTL, so this should not be
    used.

    simulate () does Monte Carlo approximation to the exact, using
    importance sampling [sampled each allele with 50% in random_allele()]
    number of iterations [n_fou_sum_MC] need to be specified by user
    (better to use more iterations for 2 QTL)

 BUGS
    simulate has not been tested yet.
    FGL_1 and FGL_0 are created in mc_twoqtl.c before calling evaluate.

 AUTHOR
    written by Yun Ju Sung

 HISTORY
    documented                           Apr 15, 2005
    updated to handle componentwise      May 19, 2005
    document revised                     Jul  2, 2005
    simulate () tested and a bug fixed   Jul 14, 2005
    extended to gene-gene interaction    Sep 19, 2006
*/

#include "twoqtl.h"

static int n_qtl = 0;
static int fou_sum_MC = -1;
static int n_fou_sum_MC = -1;
static int gxg_interaction = -1;
static int debug_level = -1;

static double allele_freq_1 = -1.0;
static double allele_freq_2 = -1.0;
static double sim_freq_1 = -1.0;
static double sim_freq_2 = -1.0;

static double *q_trait = NULL, *geno_means_1, *geno_means_2 = NULL;
static double **geno_means_joint = NULL;
static int *nFGL_c = NULL, *obs_ibase, *obs_size_c;
static int obs_size_now, comp_no_now;

static double enumerate (int **FGL_1, int **FGL_2, int *obs_idx_1,
                         int *obs_idx_2, int n_obs_1, int n_obs_2, double *y);
static double simulate (int **FGL_1, int **FGL_2, int *obs_idx_1,
                        int *obs_idx_2, int n_obs_1, int n_obs_2, double *y);

static char *next_allele (int n_obs, unsigned long int index, int *n_A_p);
static char *random_allele (int n_obs, double freq, int *n_A_p);
static void adj_qtl_effect (double *y, int **FGL, int *obs_idx, char *allele,
                            int qtl_no);
static void adj_qtl_joint (double *y, int **FGL1, int **FGL2,
                           int *obs1, int *obs2, char *allele1, char *allele2);
static double log_prob_jnt ();
static double log_prob_allele (int n_A, int n_obs, double P);
static int *set_obs_FGL_idx (int **FGL, int *size_p);
static void set_evaluate (void);

double evaluate (int **FGL_1, int **FGL_2, int comp_no)
{
   /* count the observed FGL */
   int   n_obs_1, *obs_idx_1;
   int   n_obs_2 = 0, *obs_idx_2 = NULL;
   double *y, log_prob;

   if (n_qtl == 0)
      set_evaluate ();

   if ((FGL_2 == NULL) && (n_qtl == 2))
      give_up_where (WHERE, ERR_qit);

   comp_no_now = comp_no;
   obs_size_now = obs_size_c[comp_no];

   obs_idx_1 = set_obs_FGL_idx (FGL_1, &n_obs_1);
   if (n_qtl == 2)
      obs_idx_2 = set_obs_FGL_idx (FGL_2, &n_obs_2);

   y = (double *) malloc (obs_size_now * sizeof (double));
   alloc_chk (WHERE, y);

   log_prob = 0.0;
   if (fou_sum_MC == 0)
      log_prob = enumerate (FGL_1, FGL_2, obs_idx_1, obs_idx_2,
                            n_obs_1, n_obs_2, y);
   else
      log_prob = simulate (FGL_1, FGL_2, obs_idx_1, obs_idx_2,
                           n_obs_1, n_obs_2, y);

   free (y);
   free (obs_idx_1);
   if (obs_idx_2)
      free (obs_idx_2);

   return log_prob;
}

double enumerate (int **FGL_1, int **FGL_2, int *obs_idx_1, int *obs_idx_2,
                  int n_obs_1, int n_obs_2, double *y)
{
   unsigned long int n_geno_1, n_geno_2, j, k;
   double foo_v, max_v, value;

   n_geno_1 = 1;
   n_geno_1 <<= n_obs_1;                /* n_geno = 2^(n_obs_FGL) */

   n_geno_2 = 1;
   n_geno_2 <<= n_obs_2;

   for (j = 0; j < n_geno_1; ++j)
      for (k = 0; k < n_geno_2; ++k)
      {
         value = log_prob_jnt (y, obs_idx_1, obs_idx_2, n_obs_1, n_obs_2,
                               FGL_1, FGL_2, j, k);
         update_batch (value, (j == 0) && (k == 0), &foo_v, &max_v);
      }

   if (debug_level >= 3)
      fprintf (stderr, "evaluate: %e .....\n", log (foo_v) + max_v);
   return max_v + log (foo_v);
}

double simulate (int **FGL_1, int **FGL_2, int *obs_idx_1,
                 int *obs_idx_2, int n_obs_1, int n_obs_2, double *y)
{
   double foo_v, max_v, value;
   int   iteration;

   for (iteration = 0; iteration < n_fou_sum_MC; ++iteration)
   {
      value = log_prob_jnt (y, obs_idx_1, obs_idx_2, n_obs_1, n_obs_2,
                            FGL_1, FGL_2, (unsigned long int) 0,
                            (unsigned long int) 0);
      update_batch (value, iteration == 0, &foo_v, &max_v);
   }
   return max_v + log (foo_v / n_fou_sum_MC);
}

double log_prob_jnt (double *y, int *obs_idx_1, int *obs_idx_2, int n_obs_1,
                     int n_obs_2, int **FGL_1, int **FGL_2,
                     unsigned long int j, unsigned long int k)
{
   int   n_A;
   char *allele, *allele2;
   double log_prob;

   memcpy (y, q_trait + obs_ibase[comp_no_now], obs_size_now * sizeof (double));

   if (fou_sum_MC == 0)
      allele = next_allele (n_obs_1, j, &n_A);
   else
      allele = random_allele (n_obs_1, sim_freq_1, &n_A);

   log_prob = log_prob_allele (n_A, n_obs_1, allele_freq_1);
   if (fou_sum_MC == 1)
      log_prob -= log_prob_allele (n_A, n_obs_1, sim_freq_1);

   if (gxg_interaction == 0)
      adj_qtl_effect (y, FGL_1, obs_idx_1, allele, 1);

   if (debug_level == 3)
      fprintf (stderr, "%2lx %2lx %6.4f ", j, k, exp (log_prob));

   if (n_qtl == 2)
   {
      if (fou_sum_MC == 0)
         allele2 = next_allele (n_obs_2, k, &n_A);
      else
         allele2 = random_allele (n_obs_2, sim_freq_2, &n_A);

      log_prob += log_prob_allele (n_A, n_obs_2, allele_freq_2);
      if (fou_sum_MC == 1)
         log_prob -= log_prob_allele (n_A, n_obs_2, sim_freq_2);

      if (gxg_interaction == 0)
         adj_qtl_effect (y, FGL_2, obs_idx_2, allele2, 2);
      else
         adj_qtl_joint (y, FGL_1, FGL_2, obs_idx_1, obs_idx_2, allele, allele2);

      free (allele2);
   }
   free (allele);

   log_prob += tq_gauss_quant (y, comp_no_now, 0.0, 1.0, 0);
   return log_prob;
}

char *next_allele (int n_obs, unsigned long int index, int *n_A_p)
{
   int   n_A = 0;
   int   i;
   char *allele, allele_i;

   allele = (char *) malloc (n_obs * sizeof (char));
   alloc_chk (WHERE, allele);

   for (i = 0; i < n_obs; ++i)
   {
      allele_i = (index & (1 << i)) != 0; /* bitwise */
      allele[i] = allele_i;
/*      fprintf (stderr, "%2d", (int) allele_i);*/
      n_A += !allele_i;                 /* 0 == A  and  1 == a */
   }
   *n_A_p = n_A;

   return allele;
}

char *random_allele (int n_obs, double freq, int *n_A_p)
{
   int   n_A = 0;
   int   i;
   char *allele, allele_i;

   allele = (char *) malloc (n_obs * sizeof (char));
   alloc_chk (WHERE, allele);

   for (i = 0; i < n_obs; ++i)
   {                                    /* generate A with prob freq */
      allele_i = (rans_alt () < freq) ? 0 : 1;
      allele[i] = allele_i;
/*      fprintf (stderr, "%2d", (int) allele_i);*/
      n_A += !allele_i;                 /* 0 == A  and  1 == a */
   }
   *n_A_p = n_A;

   return allele;
}

/*
 * subtract each QTL effect from trait
 */
void adj_qtl_effect (double *y, int **FGL, int *obs, char *allele, int qtl_no)
{
   int  *FGL_1 = FGL[1];
   int  *FGL_0 = FGL[0];
   int   i;
   char  genotype;

   if (debug_level == 4)
      fprintf (stderr, ".. adjust qtl effect ..\n");

   for (i = 0; i < obs_size_now; ++i)
   {
      genotype = allele[obs[FGL_0[i] - 1]];
      genotype += allele[obs[FGL_1[i] - 1]];

      if (debug_level == 4)
         fprintf (stderr, "%2d %6.4f ", i, y[i]);

      if (qtl_no == 1)
      {
         y[i] -= geno_means_1[(int) genotype];
         if (debug_level == 4)
            fprintf (stderr, "%4.2f %6.4f \n",
                     geno_means_1[(int) genotype], y[i]);
      }
      else
      {
         y[i] -= geno_means_2[(int) genotype];
         if (debug_level == 4)
            fprintf (stderr, "%4.2f %6.4f \n",
                     geno_means_2[(int) genotype], y[i]);
      }
   }
   return;
}

/*
 * subtract both QTL effects from trait (for GxG interaction)
 */
void adj_qtl_joint (double *y, int **FGL1, int **FGL2,
                    int *obs1, int *obs2, char *allele1, char *allele2)
{
   int  *FGL1_1 = FGL1[1];
   int  *FGL1_0 = FGL1[0];
   int  *FGL2_1 = FGL2[1];
   int  *FGL2_0 = FGL2[0];
   int   i;
   char  genotype1, genotype2;

   if (debug_level == 4)
      fprintf (stderr, ".. adjust qtl effect jointly..\n");

   for (i = 0; i < obs_size_now; ++i)
   {
      genotype1 = allele1[obs1[FGL1_0[i] - 1]];
      genotype1 += allele1[obs1[FGL1_1[i] - 1]];

      genotype2 = allele2[obs2[FGL2_0[i] - 1]];
      genotype2 += allele2[obs2[FGL2_1[i] - 1]];

      if (debug_level == 4)
         fprintf (stderr, "%2d %6.4f ", i, y[i]);

      y[i] -= geno_means_joint[(int) genotype1][(int) genotype2];
      if (debug_level == 4)
         fprintf (stderr, "%4.2f %6.4f \n",
                  geno_means_joint[(int) genotype1][(int) genotype2], y[i]);
   }
   return;
}

/*
 * log probability of founder genes
 */
double log_prob_allele (int n_A, int n_obs, double P)
{
   double log_prob = n_A * log (P) + (n_obs - n_A) * log (1 - P);

   if (debug_level == 5)
      fprintf (stderr, "log_prob_allele %d %d %e %e \n", n_A, n_obs, P,
               log_prob);

   return log_prob;
}

int  *set_obs_FGL_idx (int **FGL, int *size_p)
{
   int   n_FGL = nFGL_c[comp_no_now];
   int  *FGL_1 = FGL[1];
   int  *FGL_0 = FGL[0];
   int   n_observed = 0;
   int   n_obs_FGL = 0;
   int  *observed, *obs_FGL_idx, pat, mat, j;

   observed = (int *) malloc (n_FGL * sizeof (int));
   alloc_chk (WHERE, observed);
   memset (observed, 0, n_FGL * sizeof (int));

   for (j = 0; j < obs_size_now; ++j)
   {
      pat = FGL_1[j] - 1;
      mat = FGL_0[j] - 1;

      if (observed[pat] == 0)
      {
         observed[pat] = 1;
         ++n_observed;
      }
      if (observed[mat] == 0)
      {
         observed[mat] = 1;
         ++n_observed;
      }

      if (n_observed == n_FGL)
         break;
   }

   obs_FGL_idx = (int *) malloc (n_FGL * sizeof (int));
   alloc_chk (WHERE, obs_FGL_idx);

   for (j = 0; j < n_FGL; ++j)
      if (observed[j] == 1)
         obs_FGL_idx[j] = n_obs_FGL++;
      else
         obs_FGL_idx[j] = -1;           /* this is out of array index */

   if (debug_level >= 3)
   {
      fprintf (stderr, "....observed...\n");

      for (j = 0; j < n_FGL; ++j)
         fprintf (stderr, "%2d", observed[j]);
      fprintf (stderr, "  %2d  %2d \n", n_observed, n_FGL);

      for (j = 0; j < n_FGL; ++j)
         fprintf (stderr, "%2d", obs_FGL_idx[j]);
      fprintf (stderr, "  %2d\n", n_obs_FGL);
   }

   if (n_obs_FGL != n_observed)
      give_up_where (WHERE, ERR_qit);

   free (observed);

   *size_p = n_observed;
   return obs_FGL_idx;
}
void set_evaluate (void)
{
   int   j, index;

   /* TEMP */
   n_qtl = get_twoqtl_parm_int ("n_qtl");
   fou_sum_MC = get_twoqtl_parm_int ("fou_sum_MC");
   debug_level = get_twoqtl_parm_int ("debug_level");
   gxg_interaction = get_twoqtl_parm_int ("gxg_interaction");

   if (fou_sum_MC == 1)
   {
      n_fou_sum_MC = get_twoqtl_parm_int ("n_fou_sum_MC");
      sim_freq_1 = get_twoqtl_parm_dbl ("sim_freq_1");
      if (n_qtl == 2)
         sim_freq_2 = get_twoqtl_parm_dbl ("sim_freq_2");
   }

   allele_freq_1 = get_twoqtl_parm_dbl ("allele_freq_1");
   if (gxg_interaction == 0)
      geno_means_1 = get_geno_means ("geno_means_1");

   if (n_qtl == 2)
   {
      allele_freq_2 = get_twoqtl_parm_dbl ("allele_freq_2");
      if (gxg_interaction == 1)
      {
         geno_means_joint = (double **) malloc (3 * sizeof (double *));
         alloc_chk (WHERE, geno_means_joint);

         geno_means_joint[0] = get_geno_means ("geno_means_AA");
         geno_means_joint[1] = get_geno_means ("geno_means_Aa");
         geno_means_joint[2] = get_geno_means ("geno_means_aa");
      }
      else
         geno_means_2 = get_geno_means ("geno_means_2");
   }

   nFGL_c = get_nFGL_c ();
   obs_ibase = get_obs_ibase ();
   obs_size_c = get_obs_size_c ();

   if (q_trait == NULL)
   {
      /* q_trait has 0-origin-index */
      q_trait = (double *) malloc (Nq * sizeof (double));
      alloc_chk (WHERE, q_trait);       /* Nq is number of observed ind */

      for (j = index = 0; j < pedsize; ++j)
         if (nghd[j].q_dat != NULL)
            q_trait[index++] = nghd[j].q_dat[0];
   }
   return;
}

void free_evaluate (void)
{
   if (q_trait != NULL)
      free (q_trait);
   if (geno_means_joint != NULL)
      free (geno_means_joint);
}
