/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/*
 NAME 
    log_prob_alt, log_prob_null,

 SYNOPSIS

    #include "twoqtl.h"
    void log_prob_alt (double *theta, int new, 
		   char ***S_trait_1, char ***S_trait_0, int comp_no, 
		   int nsw, int debug_level, int start, mc_eval_t *mc_eval_p);
    void log_prob_null (double *theta, int new, 
                        int comp_no, int nsw, int debug_level,
		        int start, mc_eval_t *mc_eval_p);

 DESCRIPTION
    compute log Prob (traits | S_traits) for one component with 
    realized S_traits and add it to mc_eval.

 INPUT

    S_trait_1   paternal segregation indicators at trait positions      
    S_trait_0   maternal segregation indicators at trait positions      
    comp_no     0, 1, ..., n_compnt - 1
    start       1 if current batch just starts, 0 otherwise            
    debug_level level of intermediate results, 0, 1, 2, 3 
    nsw         number of meioses that need to be updated               
    mc_eval 

    theta, new are reserved for future use, not used right now.

 AUTHOR
     written by Yun Ju Sung

 HISTORY
     documented                                  Apr 15, 2005
     extend to componentwise computation         May 19, 2005
     change to realize SI's at the null locus    Jun 18, 2005
     document revised                            Jul  2, 2005

 SEE ALSO
     update_S_trait.c for allocating, realizing, and freeing S_trait.
     batch_means.c, twoqtl.h  for mc_eval

*/

#include "twoqtl.h"

static int *setup_map (char **S_1, char **S_0, int comp_no, int *size_p);
static char **pack_S (char **S, int *map, int map_size, int comp_no);

static int **set_FGL_c (char *S_1, char *S_0, int comp_no);
static int *obs_FGL (int *FGL, int comp_no);
static char **update_S_null (int comp_no);
static void get_parm (int nsw);

/* printing intermediate results for debugging */
static void print_step_1 (int *map, int map_size, char **S_1, char **S_0,
                          char **packed_S_1, char **packed_S_0, int comp_no,
                          int debug_level);
static void print_step_2 (int **FGL_1, int **FGL_0, int n_loci, int comp_no);
static void print_step_2_null (char **S_n, int **F_a, int *F_1, int *F_0,
                               int locus_no, int comp_no, int debug_level);
static void print_step_3 (double **log_prob, int map_size);

static int *ibase = NULL, *size_c, *obs_ibase, *obs_size_c, *mbase, *n_meiosis;
static int n_qtl, **obs_index;

/*
 ALGORITHM for log_prob_alt
     1. pack S (S_1, S_0) for storing only changes in S                    
        (not so much to save memory but to reduce computation later)       
     2. obtain the FGL of observed individuals at each locus in packed S.  
     3. compute Prob (Trait | FGL) with FGL at loci in packed S.           
     4. unpack S and update mc_eval                                        
*/

void log_prob_alt (double *theta, int new,
                   char ***S_trait_1, char ***S_trait_0, int comp_no,
                   int nsw, int debug_level, int start, mc_eval_t * mc_eval_p)
{
   int  *map, map_size;
   int **FGL_1, **FGL_0, **FGL_all;
   int  *FGL_a[2];                      /* obs FGL at the first locus */
   int  *FGL_b[2];                      /* obs FGL at the second locus */
   int   i, j, n_loci_1, n_loci_2, packed_i;

   double **log_prob, *log_pr;
   double *from, *to_foo, *to_max;

   char **S_1 = S_trait_1[comp_no];
   char **S_0 = S_trait_0[comp_no];
   char **packed_S_1, **packed_S_0;

   if (!ibase)
      get_parm (nsw);

   /* Pack S */
   map = setup_map (S_1, S_0, comp_no, &map_size);
   packed_S_1 = pack_S (S_1, map, map_size, comp_no);
   packed_S_0 = pack_S (S_0, map, map_size, comp_no);

   if (debug_level > 0)
      print_step_1 (map, map_size, S_1, S_0, packed_S_1, packed_S_0, comp_no,
                    debug_level);

   /* Obtain FGL of obs. ind. at each locus in packed S */

   FGL_1 = (int **) malloc (map_size * sizeof (int *));
   alloc_chk (WHERE, FGL_1);

   FGL_0 = (int **) malloc (map_size * sizeof (int *));
   alloc_chk (WHERE, FGL_0);

   for (i = 0; i < map_size; ++i)
   {
      FGL_all = set_FGL_c (packed_S_1[i], packed_S_0[i], comp_no);
      FGL_1[i] = obs_FGL (FGL_all[1], comp_no);
      FGL_0[i] = obs_FGL (FGL_all[0], comp_no);
      free_2d ((void *) &FGL_all);
   }

   if (debug_level > 3)
      print_step_2 (FGL_1, FGL_0, map_size, comp_no);

   /* Compute log probability of trait given each observed FGL */

   log_prob = alloc_dbl2d (map_size, (n_qtl == 1) ? 1 : map_size);

   for (i = 0; i < map_size; ++i)
   {
      FGL_a[1] = FGL_1[i];
      FGL_a[0] = FGL_0[i];

      if (debug_level == 3)
         fprintf (stderr, "********* %2d \n", i);

      if (n_qtl == 1)
         log_prob[i][0] = evaluate (FGL_a, NULL, comp_no);
      else
      {
         log_pr = log_prob[i];
         for (j = 0; j < map_size; ++j)
         {
            FGL_b[1] = FGL_1[j];
            FGL_b[0] = FGL_0[j];
            log_pr[j] = evaluate (FGL_a, FGL_b, comp_no);
         }
      }
   }

   if (debug_level > 0)
      print_step_3 (log_prob, map_size);

   /* Update mc_eval, while unpacking S */

   n_loci_1 = mc_eval_p->n_loci_1;      /* n_loci in the original map */
   n_loci_2 = mc_eval_p->n_loci_2;

   for (i = 0; i < n_loci_1; ++i)
   {
      packed_i = map[i];                /* locus in the packed map */
      from = log_prob[packed_i];
      to_foo = mc_eval_p->foo_alt[i];
      to_max = mc_eval_p->max_alt[i];

      for (j = 0; j < n_loci_2; ++j)
         update_batch (from[map[j]], start, to_foo + j, to_max + j);
   }

   free_2d ((void *) &log_prob);

   for (i = 0; i < map_size; ++i)
   {
      free (FGL_1[i]);
      free (FGL_0[i]);
   }
   free (FGL_1);
   free (FGL_0);

   free (map);
   free_char2d (packed_S_1);
   free_char2d (packed_S_0);

   return;
}

void log_prob_null (double *theta, int new,
                    int comp_no, int nsw, int debug_level,
                    int start, mc_eval_t * mc_eval_p)
{
   int **FGL_1, **FGL_0, **FGL_all;
   int  *FGL_a[2];                      /* obs FGL at the first locus */
   int  *FGL_b[2];                      /* obs FGL at the second locus */
   int   i;

   double log_prob;
   char **S;

   if (!ibase)
      get_parm (nsw);

   FGL_1 = (int **) malloc (n_qtl * sizeof (int *));
   alloc_chk (WHERE, FGL_1);

   FGL_0 = (int **) malloc (n_qtl * sizeof (int *));
   alloc_chk (WHERE, FGL_0);

   for (i = 0; i < n_qtl; ++i)
   {
      S = update_S_null (comp_no);
      FGL_all = set_FGL_c (S[1], S[0], comp_no);
      FGL_1[i] = obs_FGL (FGL_all[1], comp_no);
      FGL_0[i] = obs_FGL (FGL_all[0], comp_no);

      if (debug_level > 1)
         print_step_2_null (S, FGL_all, FGL_1[i], FGL_0[i],
                            i, comp_no, debug_level);

      free_2d ((void *) &FGL_all);
      free_char2d (S);
   }

   FGL_a[1] = FGL_1[0];
   FGL_a[0] = FGL_0[0];

   if (n_qtl == 1)
      log_prob = evaluate (FGL_a, NULL, comp_no);
   else
   {
      FGL_b[1] = FGL_1[1];
      FGL_b[0] = FGL_0[1];
      log_prob = evaluate (FGL_a, FGL_b, comp_no);
   }

   if (debug_level > 1)
      fprintf (stderr, "null %e\n", log_prob);

   update_batch (log_prob, start, &(mc_eval_p->foo_null),
                 &(mc_eval_p->max_null));

   for (i = 0; i < n_qtl; ++i)
   {
      free (FGL_1[i]);
      free (FGL_0[i]);
   }
   free (FGL_1);
   free (FGL_0);

   return;
}

/* map i-th locus in S to locus in packed S */
int  *setup_map (char **S_1, char **S_0, int comp_no, int *size_p)
{
   int  *map, m_start, m_stop, i, m, n_loci, change, k, j;
   char **S;

   map = (int *) malloc (NLocPos * sizeof (int));
   alloc_chk (WHERE, map);

   m_start = mbase[comp_no];
   m_stop = m_start + n_meiosis[comp_no];

   n_loci = 0;
   map[0] = n_loci;                     /* include first locus */

   for (i = 1; i < NLocPos; ++i)
   {
      change = 0;

      for (m = m_start; m < m_stop; ++m)
      {
         k = meiosis[m].sex;
         j = comp_index ((meiosis[m].ind).index, comp_no);

         S = (k == 1) ? S_1 : S_0;
         if (S[i][j] != S[i - 1][j])
         {
            change = 1;
            break;
         }
      }

      if (change)
         map[i] = ++n_loci;
      else
         map[i] = n_loci;
   }

   *size_p = ++n_loci;                  /* n_loci in packed S */

   return map;
}

char **pack_S (char **S, int *map, int map_size, int comp_no)
{
   int   comp_size = size_c[comp_no];
   int   packed_i, i, j;

   char **packed_S, *S_to, *S_from;

   packed_S = alloc_char2d (map_size, comp_size);

   packed_i = 0;
   for (i = 0; i < NLocPos; ++i)
      if (i == 0 || (map[i] != map[i - 1]))
      {
         S_to = packed_S[packed_i];
         S_from = S[i];

         for (j = 0; j < comp_size; ++j)
            S_to[j] = S_from[j];

         ++packed_i;
      }

   return packed_S;
}

int **set_FGL_c (char *S_1, char *S_0, int comp_no)
/* updated from set_FGL () in Sample*/
{
   int   comp_size = size_c[comp_no];
   int **G, *G_1, *G_0;
   int   cg, i, i_nghd, ma_index, pa_index;

   Ind  *dad, *mum;

   G = alloc_int2d (2, comp_size);
   G_1 = G[1];
   G_0 = G[0];

   cg = 0;
   for (i = 0; i < comp_size; ++i)
   {
      i_nghd = nghd_index (i, comp_no);
      dad = nghd[i_nghd].pa;

      if (S_0[i] == -1 && isNullNode (dad))
      {
         ++cg;
         G_0[i] = cg;
         ++cg;
         G_1[i] = cg;
      }
      else
      {
         mum = nghd[i_nghd].ma;
         ma_index = comp_index (mum->index, comp_no);

         if ((S_0[i] == -1) || (S_0[i] == 0))
            G_0[i] = G_0[ma_index];
         else
            G_0[i] = G_1[ma_index];

         pa_index = comp_index (dad->index, comp_no);

         if ((S_1[i] == -1) || (S_1[i] == 0))
            G_1[i] = G_0[pa_index];
         else
            G_1[i] = G_1[pa_index];
      }
   }

   return G;
}

int  *obs_FGL (int *FGL, int comp_no)
{
   int   obs_size, *o_index, *obs_FGL, j;

   obs_size = obs_size_c[comp_no];
   o_index = obs_index[comp_no];

   obs_FGL = (int *) malloc (obs_size * sizeof (int));
   alloc_chk (WHERE, obs_FGL);

   for (j = 0; j < obs_size; ++j)
      obs_FGL[j] = FGL[o_index[j]];

   return obs_FGL;
}

char **update_S_null (int comp_no)
{
   int   comp_size = size_c[comp_no];
   int   m_start, m_stop, m, k, j;
   char **S_null, *S_1, *S_0, *S;

   S_null = alloc_char2d (2, comp_size);
   S_1 = S_null[1];
   S_0 = S_null[0];

   memset (S_1, -1, comp_size * sizeof (char));
   memset (S_0, -1, comp_size * sizeof (char));

   m_start = mbase[comp_no];
   m_stop = m_start + n_meiosis[comp_no];

   for (m = m_start; m < m_stop; ++m)
   {
      k = meiosis[m].sex;
      j = comp_index ((meiosis[m].ind).index, comp_no);
      S = (k == 0) ? S_0 : S_1;
      S[j] = (rans () < .5) ? 0 : 1;
   }

   return S_null;
}

void get_parm (int nsw)
{
   ibase = get_ibase ();
   size_c = get_size_c ();

   obs_ibase = get_obs_ibase ();
   obs_size_c = get_obs_size_c ();

   mbase = get_mbase (nsw);
   n_meiosis = get_n_meiosis (nsw);

   obs_index = get_obs_index ();

   n_qtl = get_twoqtl_parm_int ("n_qtl");
   return;
}

/* ... printing functions  ... */

void print_step_1 (int *map, int map_size, char **S_1, char **S_0,
                   char **packed_S_1, char **packed_S_0,
                   int comp_no, int debug_level)
{
   int   i;

   fprintf (stderr, ".... check step 1 : component %2d .... \n", comp_no);
   if (debug_level > 1)
   {
      for (i = 0; i < NLocPos; ++i)
         fprintf (stderr, "%2d %2d\n", i, map[i]);
   }
   if (map[NLocPos - 1] != map_size - 1)
      give_up_where (WHERE, ERR_qit);

   print_S (packed_S_1, packed_S_0, map_size, comp_no);

   if (debug_level == 3)
      print_S (S_1, S_0, NLocPos, comp_no);

   return;
}

void print_step_2 (int **FGL_1, int **FGL_0, int n_loci, int comp_no)
{
   int  *FGL_1_locus, *FGL_0_locus;
   int   obs_size = obs_size_c[comp_no];
   int   locus, j;

   fprintf (stderr, ".... check step 2 : obs FGL for component %2d . \n",
            comp_no);

   for (locus = 0; locus < n_loci; ++locus)
   {
      FGL_1_locus = FGL_1[locus];
      FGL_0_locus = FGL_0[locus];

      fprintf (stderr, "locus %3d", locus);

      for (j = 0; j < obs_size; ++j)
         fprintf (stderr, " %3d%3d", FGL_1_locus[j], FGL_0_locus[j]);
      fprintf (stderr, "\n");
   }

   return;
}

void print_step_2_null (char **S_n, int **F_a, int *F_1, int *F_0,
                        int locus_no, int comp_no, int debug_level)
{
   int   i, i_nghd, obs_size, *o_index;

   fprintf (stderr, ".... check step 2 : null locus %d component %2d ... \n",
            locus_no, comp_no);

   if (debug_level == 3)
   {
      char *S_1 = S_n[1];
      char *S_0 = S_n[0];
      int  *F_a_1 = F_a[1];
      int  *F_a_0 = F_a[0];

      int  *G_1 = gg1[0];
      int  *G_0 = gg0[0];

      int   comp_size = size_c[comp_no];

      for (i = 0; i < comp_size; ++i)
      {
         fprintf (stderr, "%2d %2d  %2d %2d ", (int) S_1[i], (int) S_0[i],
                  F_a_1[i], F_a_0[i]);
         if (locus_no == 0)
         {
            i_nghd = nghd_index (i, comp_no);
            fprintf (stderr, "%2d %2d\n", G_1[i_nghd], G_0[i_nghd]);
         }
         else
            fprintf (stderr, "\n");
      }
   }

   obs_size = obs_size_c[comp_no];
   o_index = obs_index[comp_no];

   for (i = 0; i < obs_size; ++i)
      fprintf (stderr, "%2d %2d  %2d %2d\n", i, o_index[i], F_1[i], F_0[i]);

   return;
}

void print_step_3 (double **log_prob, int map_size)
{
   int   j_stop, i, j;
   double *log_pr;

   j_stop = (n_qtl == 1) ? 1 : map_size;
   fprintf (stderr, ".... check step 3 : log probability  ... \n");

   for (i = 0; i < map_size; ++i)
   {
      log_pr = log_prob[i];

      fprintf (stderr, "%2d ", i);
      for (j = 0; j < j_stop; ++j)
         fprintf (stderr, "%e ", log_pr[j]);
      fprintf (stderr, "\n");
   }

   return;
}
