/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/*
 NAME
     peel_poly, sample_poly, tq_gauss_quant, free_tq_gauss

 SYNOPSIS

     #include "twoqtl.h"
     void peel_poly (double *y, double *log_prob);
     void sample_poly (double *y, double *poly);
     double tq_gauss_quant (double *y_comp, int comp_no,
                            double sigd, double lambda, int im);
     void free_tq_gauss (void);

 DESCRIPTION

     peel_poly computes log probability of the observed trait data y
     for each component.

     sample_poly samples polygenic values of all individuals in pedigree,
     conditional on y.

     tq_gauss_quant computes log probability of y_comp for one component.

     free_tq_gauss frees all the static data structures in tq_gauss_quant.

 INPUT

     y         mean-adjusted quantitative trait (of length Nq)
               (0-origin indexing, differing from gauss_qaunt in Peel)

     log_prob  array of length n_compnts for log prob
     poly      array of length pedsize for polygenic values

     y_comp    mean-adjusted quantitative trait for one component
               (of length number of observed ind. in this component)
     comp_no   0, 1, ..., n_compnts - 1

     sigd, lambda and im  are copied from gauss_quant in Peel and
     not completely understood (see DETAILS, REFERENCE, and BUGS below).
     lm_twoqtl calls tq_gauss_quant with sigd = 0.0, lambda = 1.0, im = 0,
     which corresponds to the additive model for polygenic values.

 GLOBAL VARIABLES USED
     n_compnts, nghd,

 DETAILS

     peel_poly and sample_poly call tq_gauss_quant for each component
     with other parameters set to sigd = 0.0, lambda = 1.0, im = 0.

     tq_gauss_quant() is updated from gauss_quant() (as of 05/08/2005).
     (1) computes likelihood for each component of pedigree.
     (2) malloc/free is simplified.
     (3) assumes nghd is ordered (offspring comes after their parents)
         within each component and individuals in the same component are
         together
         ---otherwise, it complains and stops.
     (4) use of global qdex is removed.
     (5) change y into 0-origin indexing (from 1-origin indexing).

 REFERENCE
     Parameter lamdba is described by
     Monte Carlo likelihood in the genetic mapping of complex traits
     E A Thompson, Phil Trans R Soc Lond B (1994) 344, 345--351.

 BUGS
    SigErr (env. var.) and SigAdd (add. var.) are not set properly
    except for lm_twoqtl.

    tq_gauss_quant has not been tested except with sigd = 0.0, lambda = 1.0,
    im = 0.

 AUTHOR
    Written by Yun Ju Sung

 Document revised  Jul 2, 2005.
*/

#include "twoqtl.h"

/* Sparse matrix storage off-diagonal element. */
struct Off
{
   struct Off *Next;
   int   col;
   double val;
};

/* Sparse matrix storage diagonal element. */
struct Diag
{
   struct Off *First;
   double val;
};

static struct Diag *BMatrix = NULL;

static int *obs_ibase = NULL;           /* index of first obs ind of comp */
static int *ibase = NULL;
static int *size_c = NULL;              /* number of ind in comp  */
static int debug_level = 0;             /* TEMP */

static void sample_poly_comp (double *y, double *poly, int comp_no);
static void print_BMatrix (int comp_no);

/*............... main function 1      ...................*/
void peel_poly (double *y, double *log_prob)
{
   int   i;

   if (!obs_ibase)
      obs_ibase = get_obs_ibase ();

   for (i = 0; i < n_compnts; i++)
      log_prob[i] = tq_gauss_quant (y + obs_ibase[i], i, 0.0, 1.0, 0);

   if (debug_level == 2)
      test_tq_gauss_quant (y, log_prob);

   free_tq_gauss ();

   return;
}

/*............... main function 2      ...................*/
void sample_poly (double *y, double *poly)
{
   int   i;

   if (!ibase)
   {
      ibase = get_ibase ();
      obs_ibase = get_obs_ibase ();
   }
   for (i = 0; i < n_compnts; i++)
      sample_poly_comp (y + obs_ibase[i], poly + ibase[i], i);

   free_tq_gauss ();

   return;
}

/* Sample polygenic values by component. */
void sample_poly_comp (double *y, double *z, int comp_no)
{
   int   comp_size = size_c[comp_no];
   int   i, i_BMatrix;
   double mean, variance;
   struct Off *p;

   if (debug_level == 2)
      fprintf (stderr, "*..... sample poly in component %d\n", comp_no);

   tq_gauss_quant (y, comp_no, 0.0, 1.0, 0); /* Only additive genetic part.   */

   for (i = 0; i < comp_size; ++i)
   {
      /* Conditional mean and variance of Z_i given y and z_0 up to z_(i-1).  */
      mean = 0;
      i_BMatrix = i + 1;                /* First index in BMatrix is for constant. */
      variance = 1 / BMatrix[i_BMatrix].val;

      for (p = BMatrix[i_BMatrix].First; p != NULL; p = p->Next)
         mean -= (p->val) * variance * ((p->col == 0) ? 1.0 : z[p->col - 1]);

      z[i] = normal (mean, sqrt (variance));

      if (debug_level == 2)
         fprintf (stderr, "poly %2d %e %e %e\n", i, mean, variance, z[i]);
   }

   if (debug_level == 2)
      print_BMatrix (comp_no);

   return;
}

/*............... MODIFIED gauss_quant ...................*/

/*
 * Sparse matrix element (Numerical Recipes, Press et al.).  Only useful if you
 * know the structure of the matrix (with regard to non-zero elements) before
 * you start.  Used here to store the inverse of the G matrix.
 */
struct sp_element
{
   int   x;
   double val;
};

/* For building up contribution from an individual. */
struct rvec
{
   int   i;
   double val;
};

/*...  Componentwise parameters  ... */
static struct sp_element **AIMatrix_c;  /* inverse of nrm */
static int **data_index_c;              /* index of observed data for each ind        */
static int **eq_pos_c;
static int *nfam_c;                     /* number of sib-ships if (dom)               */
static int **fam_list_c;                /* sib-ships if (dom) and NULL otherwise      */
static int *obs_size_c = NULL;

static int BMatrix_size = 0;

static struct sp_element *tq_calc_nrm (int comp_no, int comp_size);
static void cmp_gauss_setup (int comp_no, int dom);
static void tq_gauss_setup (int dom);

struct Off *newOff (void);
static void AddElement (int r, int c, double x);
static double envonly_comp (double *y, int comp_no);
static void free_BMatrix (void);
static void realloc_BMatrix (int size);

/*
 * Calculate likelihood for polygenic trait by integrating out breeding
 * values.  Can also fit pseudo-dominance effect (common environment).
 * Designed to act as a clone of peel_quant(), except with the ability
 * to handle general matrices.
 */
double tq_gauss_quant (double *y, int comp_no,
                       double sigd, double lambda, int im)
{
   int   i, i1, j, l, t, nneg, size, nfam, dom = 0;
   int   comp_size, *data_index, *fam_list, *eq_pos;
   double k, z, nn1, v, lam, c, dt;

   struct sp_element *AIMatrix;
   struct Off *p, *p1, *op;
   struct rvec rvec[3];

   SigErr = get_twoqtl_parm_dbl ("SigErr");
   SigAdd = get_twoqtl_parm_dbl ("SigAdd");

   if (SigAdd + sigd <= 0.0)
   {
      if (obs_size_c == NULL)
         obs_size_c = get_obs_size_c ();
      return envonly_comp (y, comp_no);
   }
   lam = 1.0;
   if ((im == 1) || (im > 2))
      lam = lambda;

   if (sigd > 0.0)
      dom = 1;

   /* If first time through, set up data structures. */
   if (!size_c)
      tq_gauss_setup (dom);

   comp_size = size_c[comp_no];
   AIMatrix = AIMatrix_c[comp_no];
   data_index = data_index_c[comp_no];
   fam_list = fam_list_c[comp_no];
   eq_pos = eq_pos_c[comp_no];

   if (dom && (fam_list == NULL))
      give_up_where (WHERE, ERP_swc);

   /* No. equations. */
   size = 1 + comp_size + nfam_c[comp_no];

   realloc_BMatrix (size);

   /* Add in contributions from data. */
   nn1 = k = 0.0;
   c = SigErr + 0.75 * sigd;
   rvec[0].i = 0;
   for (i = 0; i < comp_size; i++)
   {
      j = data_index[i];
      if (j > -1)                       /* observed */
      {
         nn1++;
         rvec[0].val = -y[j];
         rvec[1].i = eq_pos[i];
         rvec[1].val = lam;
         if (dom)
         {
            j = fam_list[i] + comp_size - 1;
            rvec[2].i = eq_pos[j];
            rvec[2].val = 1.0;
            i1 = 3;
         }
         else
            i1 = 2;
         for (j = 0; j < i1; j++)
         {
            l = rvec[j].i;
            z = rvec[j].val;
            for (t = 0; t < j; t++)
               AddElement (l, rvec[t].i, z * rvec[t].val / c);
            BMatrix[l].val += z * z / c;
         }
      }
   }
   k -= .5 * nn1 * log (PPI * c);

   if (debug_level == 2)
      print_BMatrix (comp_no);

   /* Add in contributions from additive effect. */
   if (SigAdd > 0.0)
   {
      for (i = 0; i < comp_size; i++)
      {
         i1 = eq_pos[i];
         BMatrix[i1].val += AIMatrix[i].val / SigAdd;
         for (j = AIMatrix[i].x; j < AIMatrix[i + 1].x; j++)
            AddElement (i1, eq_pos[AIMatrix[j].x], AIMatrix[j].val / SigAdd);
      }
      k -= .5 * (log (PPI * SigAdd) * (double) comp_size
                 + AIMatrix[comp_size].val);
   }

   if (debug_level == 2)
      print_BMatrix (comp_no);

   /* Add in contributions from "dominance" effect. */
   if (dom)
   {
      nfam = nfam_c[comp_no];
      /* Note we are actually using sigd/4 */
      nn1 = 0.0;
      for (i = 0; i < nfam; i++)
      {
         if (BMatrix[eq_pos[i + comp_size]].val > 0.0)
         {
            BMatrix[eq_pos[i + comp_size]].val += 4.0 / sigd;
            nn1 += 1.0;
         }
      }
      k -= .5 * nn1 * (log (0.5 * M_PI * sigd));

      if (debug_level > 0)
         print_BMatrix (comp_no);
   }

   /* Now start absorption. */
   nneg = 0;
   dt = nn1 = 0.0;
   for (i = size - 1; i > 0; i--)
   {
      v = BMatrix[i].val;
      if (fabs (v) < 1.0e-16)
         continue;
      for (p = BMatrix[i].First; p != NULL; p = p->Next)
      {
         for (p1 = BMatrix[i].First; p1 != p; p1 = p1->Next)
            AddElement (p->col, p1->col, -p->val * p1->val / v);

         BMatrix[p->col].val -= p->val * p->val / v;

         op = p;
      }

      if (v < 0.0)
      {
         v = -v;
         give_warn (ERP_npv, "gauss_quant", i);
         nneg++;
      }
      dt += log (v);
      nn1 += 1.0;
   }

   k += .5 * (nn1 * log (PPI) - dt - BMatrix[0].val);

   if (nneg)
      give_warn (ERP_nnp, "gauss_quant", nneg);

   return k;
}

void realloc_BMatrix (int size)
{
   int   i;

   if (BMatrix != NULL)
      free_BMatrix ();

   /* Allocate diagonal elements. */
   BMatrix = malloc (size * sizeof (struct Diag));
   alloc_chk (WHERE, BMatrix);

   for (i = 0; i < size; i++)
   {
      BMatrix[i].val = 0.0;
      BMatrix[i].First = NULL;
   }

   BMatrix_size = size;
   return;
}

void free_BMatrix (void)
{
   int   i, size;
   struct Off *p, *p1;

   size = BMatrix_size;

   for (i = 0; i < size; i++)
   {
      p = BMatrix[i].First;
      while (p)
      {
         p1 = p;
         p = p->Next;
         free (p1);
      }
   }
   free (BMatrix);
   BMatrix = NULL;

   return;
}

void print_BMatrix (int comp_no)
{
   int   i;
   struct Off *p;

   int   size = 1 + size_c[comp_no] + nfam_c[comp_no];

   for (i = 0; i < size; i++)
   {
      for (p = BMatrix[i].First; p != NULL; p = p->Next)
         fprintf (stderr, "BMatrix %d %d %e\n", i, p->col, p->val);
      fprintf (stderr, "BMatrix %d %d %e\n", i, i, BMatrix[i].val);
   }
   return;
}

/* Calculate inverse of NRM (G) matrix using Quaas/Henderson method.  See
 * Numerical Recipes in C 2nd edition pps. 78-79 for a description of the
 * sparse matrix storage.  Note the matrix is half stored.
 *
 * comp_size by comp_size for component comp_no with size comp_size
 */
struct sp_element *tq_calc_nrm (int comp_no, int comp_size)
{
   struct sp_element *AIMatrix;

   int   i, j, k, ids, idd, ids1, idd1, nz;
   double *u, *v, xx, d, dt = 0.0, d2, d4;
   Marlist *ml;

/* Count non-zero (off diagonal) elements in lower triangular part of matrix. */
   for (i = nz = 0; i < comp_size; i++)
   {
      j = nghd_index (i, comp_no);
      ids = nghd[j].pa->index;
      idd = nghd[j].ma->index;

      nz += (ids != INVALID_INDEX);
      nz += (idd != INVALID_INDEX);
      for (ml = nghd[j].marriages; ml != NULL; ml = ml->next)
      {
         k = ml->current->spouse->index;
         if (k < j)
            nz++;
      }
   }
   AIMatrix = malloc ((nz + 1 + comp_size) * sizeof (struct sp_element));
   alloc_chk (WHERE, AIMatrix);

   /* Set up coordinates for non-zero elements. */
   nz = comp_size + 1;
   for (i = 0; i < comp_size; i++)
   {
      AIMatrix[i].val = 0.0;
      AIMatrix[i].x = nz;

      j = nghd_index (i, comp_no);
      ids = nghd[j].pa->index;
      idd = nghd[j].ma->index;

      if (ids != INVALID_INDEX)
      {
         AIMatrix[nz].x = comp_index (ids, comp_no);
         AIMatrix[nz++].val = 0.0;
      }
      if (idd != INVALID_INDEX)
      {
         AIMatrix[nz].x = comp_index (idd, comp_no);
         AIMatrix[nz++].val = 0.0;
      }

      for (ml = nghd[j].marriages; ml != NULL; ml = ml->next)
      {
         k = ml->current->spouse->index;
         if (k < j)
         {
            AIMatrix[nz].x = comp_index (k, comp_no);
            AIMatrix[nz++].val = 0.0;
         }
      }
   }
   AIMatrix[comp_size].x = nz;

   u = malloc (comp_size * 2 * sizeof (double));
   alloc_chk (WHERE, u);

   v = u + comp_size;
   for (i = 0; i < comp_size; i++)
      u[i] = 0.0;

   /* Build up matrix */
   for (i = 0; i < comp_size; i++)
   {
      xx = 0.0;

      j = nghd_index (i, comp_no);
      ids = comp_index (nghd[j].pa->index, comp_no);
      idd = comp_index (nghd[j].ma->index, comp_no);

      if (ids >= 0)
         xx = u[ids];
      if (idd >= 0)
         xx += u[idd];
      xx = 1.0 - .25 * xx;
      d = 1.0 / xx;
      u[i] += xx;
      xx = sqrt (xx);
      v[i] = xx;
      dt += log (xx);
      for (k = i + 1; k < comp_size; k++)
      {
         xx = 0.0;

         j = nghd_index (k, comp_no);
         ids1 = comp_index (nghd[j].pa->index, comp_no);
         idd1 = comp_index (nghd[j].ma->index, comp_no);

         if (ids1 >= i)
            xx = v[ids1];
         if (idd1 >= i)
            xx += v[idd1];
         if (xx > 0.0)
         {
            xx *= .5;
            u[k] += xx * xx;
         }
         v[k] = xx;
      }
      AIMatrix[i].val += d;
      d2 = -.5 * d;
      d4 = .25 * d;
      j = AIMatrix[i].x;
      if (ids >= 0)
      {
         AIMatrix[j++].val += d2;
         AIMatrix[ids].val += d4;
      }
      if (idd >= 0)
      {
         AIMatrix[j++].val += d2;
         AIMatrix[idd].val += d4;
         if (ids >= 0)
         {
            if (ids == idd)
               AIMatrix[ids].val += d4;
            else if (ids > idd)
            {
               for (j = AIMatrix[ids].x; j < AIMatrix[ids + 1].x; j++)
                  if (AIMatrix[j].x == idd)
                  {
                     AIMatrix[j].val += d4;
                     break;
                  }
            }
            else
            {
               for (j = AIMatrix[idd].x; j < AIMatrix[idd + 1].x; j++)
                  if (AIMatrix[j].x == ids)
                  {
                     AIMatrix[j].val += d4;
                     break;
                  }
            }
         }
      }
   }
   dt += dt;
   /* Log determinant of the G matrix. */
   AIMatrix[comp_size].val = dt;
   free (u);
   u = NULL;

   return AIMatrix;
}

/*
 * Setup data structures for gauss_quant().
 *       Only called once [by gauss_quant()].
 */
void cmp_gauss_setup (int comp_no, int dom)
{
   int   i, j, k, index, nfam;
   int  *data_index, *fam_list, *eq_pos;
   int   comp_size = size_c[comp_no];

   Marlist *ml;
   Indlist *kid;

   /* Calculate inverse of NRM (G) matrix. */
   AIMatrix_c[comp_no] = tq_calc_nrm (comp_no, comp_size);

   data_index_c[comp_no] = (int *) malloc (comp_size * sizeof (int));
   alloc_chk (WHERE, data_index_c[comp_no]);
   data_index = data_index_c[comp_no];

   index = 0;
   for (i = 0; i < comp_size; i++)
   {
      j = nghd_index (i, comp_no);

      if (nghd[j].q_dat != NULL)
         data_index[i] = index++;
      else
         data_index[i] = -1;
   }

   nfam = 0;
   if (dom)
   {
      /*
       * If dominance effect required, count how many levels of effect
       * we need.
       */
      fam_list_c[comp_no] = calloc (comp_size, sizeof (int));
      alloc_chk (WHERE, fam_list_c[comp_no]);
      fam_list = fam_list_c[comp_no];

      for (i = 0; i < comp_size; i++)
      {
         j = nghd_index (i, comp_no);
         for (ml = nghd[j].marriages; ml != NULL; ml = ml->next)
         {
            if (ml->current->spouse->index < j)
            {
               nfam++;
               for (kid = ml->current->kids; kid != NULL; kid = kid->next)
               {
                  k = comp_index (kid->current->index, comp_no);
                  fam_list[k] = nfam;
               }
            }
         }
      }
      for (i = 0; i < comp_size; i++)
         if (!fam_list[i] && (data_index[i] != -1))
            fam_list[i] = ++nfam;

      /*
       * Set up order of equations so the dominance effect for a family
       * is eliminated after the last offspring from that family.
       */
      eq_pos_c[comp_no] = calloc (comp_size + nfam, sizeof (int));
      alloc_chk (WHERE, eq_pos_c[comp_no]);
      eq_pos = eq_pos_c[comp_no];

      k = 0;
      for (i = 0; i < comp_size; i++)
      {
         j = fam_list[i];
         if (j != 0)
         {
            j += comp_size - 1;
            if (eq_pos[j] == 0)
               eq_pos[j] = ++k;
         }
         eq_pos[i] = ++k;
      }
   }
   else
   {
      fam_list_c[comp_no] = NULL;

      eq_pos_c[comp_no] = calloc (comp_size, sizeof (int));
      alloc_chk (WHERE, eq_pos_c[comp_no]);

      eq_pos = eq_pos_c[comp_no];
      for (i = 0; i < comp_size; i++)
         eq_pos[i] = i + 1;
   }
   nfam_c[comp_no] = nfam;

   return;
}

void tq_gauss_setup (int dom)
{
   int   i;

   if (!ibase)
      ibase = get_ibase ();

   size_c = get_size_c ();

   AIMatrix_c = (struct sp_element **)
      malloc (n_compnts * sizeof (struct sp_element *));
   alloc_chk (WHERE, AIMatrix_c);

   data_index_c = (int **) malloc (n_compnts * sizeof (int *));
   alloc_chk (WHERE, data_index_c);

   eq_pos_c = (int **) malloc (n_compnts * sizeof (int *));
   alloc_chk (WHERE, eq_pos_c);

   fam_list_c = (int **) malloc (n_compnts * sizeof (int *));
   alloc_chk (WHERE, fam_list_c);

   nfam_c = (int *) malloc (n_compnts * sizeof (int));
   alloc_chk (WHERE, nfam_c);

   for (i = 0; i < n_compnts; i++)
   {
      cmp_gauss_setup (i, dom);

      if (debug_level == 2)
         fprintf (stderr, "Component %2d  ibase = %2d size = %2d\n",
                  i, ibase[i], size_c[i]);
   }

   return;
}

/* malloc new Off */
struct Off *newOff (void)
{
   struct Off *p = (struct Off *) malloc (sizeof (struct Off));
   alloc_chk (WHERE, p);

   return p;
}

/*
 * Add x to off diagonal element r,c.  If element does not exist yet,
 * calls AllocElement() to allocate storage.
 */
void AddElement (int r, int c, double x)
{
   int   temp;
   struct Off *p, **p1, *p2;

   /* Symmetric matrix, so r should be > c. */
   if (c > r)
   {
      temp = c;
      c = r;
      r = temp;
   }
   p1 = &BMatrix[r].First;
   p = BMatrix[r].First;
   if (p != NULL)
   {
      /* If we have off-diagonal elements for this row, search through them.  */
      while (p->col < c && p->Next)
      {
         p1 = &p->Next;
         p = p->Next;
      }

      /* Element found - add x to value */
      if (p->col == c)
         p->val += x;
      else
      {
         /* Element not found - allocate new space for it. */
         p2 = newOff ();
         p2->val = x;
         p2->col = c;

         /* Link new element into list. */
         if (p->col > c)
         {
            p2->Next = p;
            (*p1) = p2;
         }
         else
         {
            p2->Next = p->Next;
            p->Next = p2;
         }
      }
   }
   else
   {
      /* No off-diagonals for this row yet, so create one. */
      p2 = newOff ();
      (*p1) = p2;
      p2->val = x;
      p2->Next = NULL;
      p2->col = c;
   }

   return;
}

/* Modified from envonly () to compute likelihood for each component. */
double envonly_comp (double *y, int comp_no)
{
   int   obs_size = obs_size_c[comp_no];

   int   i;
   double log_prob = 0;

   double SigErr = get_twoqtl_parm_dbl ("SigErr");

   for (i = 0; i < obs_size; ++i)
      log_prob += y[i] * y[i];

   log_prob /= SigErr;
   log_prob += obs_size * log (PPI * SigErr);
   log_prob *= -0.5;

   return log_prob;
}

void free_tq_gauss (void)
{
   int   i;

   free_BMatrix ();

   for (i = 0; i < n_compnts; ++i)
   {
      free (AIMatrix_c[i]);
      free (data_index_c[i]);
      free (eq_pos_c[i]);
      if (nfam_c[i] != 0)
         free (fam_list_c[i]);
   }
   free (AIMatrix_c);
   free (data_index_c);
   free (eq_pos_c);
   free (fam_list_c);
   free (nfam_c);

   return;
}
