/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/*
 NAME 
    set_twoqtl_parm_int, set_twoqtl_parm_dbl, set_geno_means,
    get_twoqtl_parm_int, get_twoqtl_parm_dbl, get_geno_means,
    free_twoqtl_parm,

 SYNOPSIS 

    #include "twoqtl.h"

    void set_twoqtl_parm_int (char *key, int value);
    void set_twoqtl_parm_dbl (char *key, double value);
    void set_geno_means (char *key, double *value);

    int    get_twoqtl_parm_int (char *key);
    double get_twoqtl_parm_dbl (char *key);
    double *get_geno_means (char *key);

    void free_twoqtl_parm (void);
 
 DESCRIPTION

    Set functions need to be called by Myrna to process parameter statements.
    Get functions get these parameters.

 AUTHOR 
    Written by Yun Ju Sung

 HISTORY        
    documented                                       Jul  1, 2005
    make default values for some parameters          Aug 25, 2005
*/

/*
  default values 

  n_batch = 20
  ScorIt = 10
  null_it = 10
  sim_freq_1 = allele_freq_1
  sim_freq_2 = allele_freq_2
  debug_level = 0
 */

#include "twoqtl.h"

static double sigerr = -1.0;            /* residual variance         */
static double sigadd = -1.0;            /* additive genetic variance */

static double allele_freq_1 = -1.0;     /* allele freq. of A in QTL 1 */
static double allele_freq_2 = -1.0;     /* allele freq. of A in QTL 2 */

static double *geno_means_1 = NULL;     /* geno means AA, Aa, aa QTL 1 */
static double *geno_means_2 = NULL;     /* geno means AA, Aa, aa QTL 2 */

static double *position_1 = NULL;
static double *position_2 = NULL;
static double *position_3 = NULL;

static double *geno_means_AA = NULL;    /* geno means for interaction */
static double *geno_means_Aa = NULL;
static double *geno_means_aa = NULL;

static double sim_freq_1 = -1.0;        /* sim. allele freq. of A in QTL 1 */
static double sim_freq_2 = -1.0;        /* sim. allele freq. of A in QTL 2 */

static double fix_qtl2_at = -1.0;       /* location of fixed QTL */
static double pos_qtl2 = -1.0;          /* location from the left marker */
static int l_marker_qtl2 = -1;          /* index of left marker for fixed qtl */
static double pos_qtl1 = -1.0;          /* location from the left marker */
static int l_marker_qtl1 = -1;          /* index of left marker for fixed qtl */
static int n_qtl = -1;                  /* number of QTL      */
static int n_batch = -1;                /* number of batches  */
static int batch_size = -1;             /* size of each batch */
static int null_it = -1;                /* spacing of MCMC for computing */

static int fou_sum_MC = -1;             /* MC(1)/exact(0) summation for FG allele */
static int gxg_interaction = -1;        /* MC(1)/exact(0) interaction */
static int n_fou_sum_MC = -1;           /* number of MC realizations for FG allele */
static int debug_level = -1;            /* debug level */

void set_twoqtl_parm_dbl (char *key, double value)
{
   if (strncmp (key, "SigErr", 20) == 0)
   {
      if (value <= 0.0)
         give_up (ERR_err, "Environmental variance must be positive\n");
      sigerr = value;
      return;
   }

   if (strncmp (key, "SigAdd", 20) == 0)
   {
      if (value < 0.0)
         give_up (ERR_err, "Additive genetic variance must be non-negative\n");
      sigadd = value;
      return;
   }

   if (strncmp (key, "allele_freq_1", 20) == 0)
   {
      if (value <= 0.0 || value >= 1.0)
         give_up (ERR_err,
                  "Allele frequency of QTL 1 must be between 0 and 1\n");
      allele_freq_1 = value;
      return;
   }

   if (strncmp (key, "allele_freq_2", 20) == 0)
   {
      if (value <= 0.0 || value >= 1.0)
         give_up (ERR_err,
                  "Allele frequency of QTL 2 must be between 0 and 1\n");
      allele_freq_2 = value;
      return;
   }

   if (strncmp (key, "sim_freq_1", 20) == 0)
   {
      if (value <= 0.0 || value >= 1.0)
         give_up (ERR_err,
                  "Simulation allele freq. for QTL 1 must be between 0 and 1\n");
      sim_freq_1 = value;
      return;
   }

   if (strncmp (key, "sim_freq_2", 20) == 0)
   {
      if (value <= 0.0 || value >= 1.0)
         give_up (ERR_err,
                  "Simulation allele freq. for QTL 2 must be between 0 and 1\n");
      sim_freq_2 = value;
      return;
   }

   if (strncmp (key, "fix_qtl2_at", 20) == 0)
   {
      fix_qtl2_at = value;
      return;
   }

   if (strncmp (key, "pos_qtl1", 20) == 0)
   {
      pos_qtl1 = value;
      return;
   }

   if (strncmp (key, "pos_qtl2", 20) == 0)
   {
      pos_qtl2 = value;
      return;
   }

   give_up (ERR_err, "set_twoqtl_parm_dbl: key not recognized\n");
}

double get_twoqtl_parm_dbl (char *key)
{
   if (strncmp (key, "SigErr", 20) == 0)
   {
      if (sigerr < 0.0)
         set_twoqtl_parm_dbl ("SigErr", PTraitPars->resi_var);
      return sigerr;
   }

   if (strncmp (key, "SigAdd", 20) == 0)
   {
      if (sigadd < 0.0)
         set_twoqtl_parm_dbl ("SigAdd", PTraitPars->addi_var);
      return sigadd;
   }

   if (strncmp (key, "pos_qtl1", 20) == 0)
   {
      if (pos_qtl1 >= 0.0)
         return pos_qtl1;
      else
         give_up (ERR_err, "pos_qtl1 not set\n");
   }

   if (strncmp (key, "pos_qtl2", 20) == 0)
   {
      if (pos_qtl2 >= 0.0)
         return pos_qtl2;
      else
         give_up (ERR_err, "pos_qtl2 not set\n");
   }

   if (strncmp (key, "allele_freq_1", 20) == 0)
   {
      if (allele_freq_1 >= 0.0 && allele_freq_1 <= 1.0)
         return allele_freq_1;
      else
         give_up (ERR_err, "allele_freq_1 not set\n");
   }

   if (strncmp (key, "allele_freq_2", 20) == 0)
   {
      if (allele_freq_2 >= 0.0 && allele_freq_2 <= 1.0)
         return allele_freq_2;
      else
         give_up (ERR_err, "allele_freq_2 not set\n");
   }

   if (strncmp (key, "fix_qtl2_at", 20) == 0)
   {
      return fix_qtl2_at;
   }

   if (strncmp (key, "sim_freq_1", 20) == 0)
   {
      if (sim_freq_1 <= 0.0 || sim_freq_1 >= 1.0)
      {
         fprintf (stdout,
                  " sim_freq_1 was set %3.2f (= allele_freq_1) by default\n",
                  allele_freq_1);
         set_twoqtl_parm_dbl ("sim_freq_1", allele_freq_1);
      }
      return sim_freq_1;
   }

   if (strncmp (key, "sim_freq_2", 20) == 0)
   {
      if (sim_freq_2 <= 0.0 || sim_freq_2 >= 1.0)
      {
         fprintf (stdout,
                  " sim_freq_2 was set %3.2f (= allele_freq_2) by default\n",
                  allele_freq_2);
         set_twoqtl_parm_dbl ("sim_freq_2", allele_freq_2);
      }
      return sim_freq_2;
   }

   give_up (ERR_err, "get_twoqtl_parm_dbl: key not recognized\n");
   return 0.0;
}

void set_position (char *key, double *value)
{
   if (strncmp (key, "position_1", 20) == 0)
   {
      position_1 = malloc (2 * sizeof (double));
      alloc_chk (WHERE, position_1);

      memcpy (position_1, value, 2 * sizeof (double));
      return;
   }
   if (strncmp (key, "position_2", 20) == 0)
   {
      position_2 = malloc (2 * sizeof (double));
      alloc_chk (WHERE, position_2);

      memcpy (position_2, value, 2 * sizeof (double));
      return;
   }
   if (strncmp (key, "position_3", 20) == 0)
   {
      position_3 = malloc (2 * sizeof (double));
      alloc_chk (WHERE, position_3);

      memcpy (position_3, value, 2 * sizeof (double));
      return;
   }
}

double *get_position (char *key)
{
   if (strncmp (key, "position_1", 20) == 0)
   {
      if (position_1 != NULL)
         return position_1;
      else
         give_up (ERR_err, "position_1 not set\n");
   }

   if (strncmp (key, "position_2", 20) == 0)
   {
      if (position_2 != NULL)
         return position_2;
      else
         give_up (ERR_err, "position_2 not set\n");
   }

   if (strncmp (key, "position_3", 20) == 0)
   {
      if (position_3 != NULL)
         return position_3;
      else
         give_up (ERR_err, "position_3 not set\n");
   }

   give_up (ERR_err, "get_position: key not recognized\n");
   return NULL;
}

void set_geno_means (char *key, double *value)
{
   if (strncmp (key, "geno_means_1", 20) == 0)
   {
      geno_means_1 = malloc (3 * sizeof (double));
      alloc_chk (WHERE, geno_means_1);

      memcpy (geno_means_1, value, 3 * sizeof (double));
      return;
   }

   if (strncmp (key, "geno_means_2", 20) == 0)
   {
      geno_means_2 = malloc (3 * sizeof (double));
      alloc_chk (WHERE, geno_means_2);

      memcpy (geno_means_2, value, 3 * sizeof (double));
      return;
   }

   if (strncmp (key, "geno_means_AA", 20) == 0)
   {
      geno_means_AA = malloc (3 * sizeof (double));
      alloc_chk (WHERE, geno_means_AA);

      memcpy (geno_means_AA, value, 3 * sizeof (double));
      return;
   }
   if (strncmp (key, "geno_means_Aa", 20) == 0)
   {
      geno_means_Aa = malloc (3 * sizeof (double));
      alloc_chk (WHERE, geno_means_Aa);

      memcpy (geno_means_Aa, value, 3 * sizeof (double));
      return;
   }

   if (strncmp (key, "geno_means_aa", 20) == 0)
   {
      geno_means_aa = malloc (3 * sizeof (double));
      alloc_chk (WHERE, geno_means_aa);

      memcpy (geno_means_aa, value, 3 * sizeof (double));
      return;
   }

   give_up (ERR_err, "set_geno_means: key not recognized\n");
}

double *get_geno_means (char *key)
{
   if (strncmp (key, "geno_means_1", 20) == 0)
   {
      if (geno_means_1 == NULL)
         set_geno_means ("geno_means_1", PTraitPars->means);
      return geno_means_1;
   }

   if (strncmp (key, "geno_means_2", 20) == 0)
   {
      if (geno_means_2 != NULL)
         return geno_means_2;
      else
         give_up (ERR_err, "geno_means_2 not set\n");
   }

   if (strncmp (key, "geno_means_AA", 20) == 0)
   {
      if (geno_means_AA != NULL)
         return geno_means_AA;
      else
         give_up (ERR_err, "geno_mean_AA not set\n");
   }

   if (strncmp (key, "geno_means_Aa", 20) == 0)
   {
      if (geno_means_Aa != NULL)
         return geno_means_Aa;
      else
         give_up (ERR_err, "geno_means_Aa not set\n");
   }

   if (strncmp (key, "geno_means_aa", 20) == 0)
   {
      if (geno_means_aa != NULL)
         return geno_means_aa;
      else
         give_up (ERR_err, "geno_means_aa not set\n");
   }

   give_up (ERR_err, "get_geno_means: key not recognized\n");
   return NULL;
}

void set_twoqtl_parm_int (char *key, int value)
{
   if (strncmp (key, "n_qtl", 20) == 0)
   {
      if (value != 1 && value != 2)
         give_up (ERR_err, "Number of QTL must be either 1 or 2\n");
      n_qtl = value;
      return;
   }

   if (strncmp (key, "n_batch", 20) == 0)
   {
      if (value <= 0)
         give_up (ERR_err, "Number of batches must be positive\n");
      n_batch = value;
      return;
   }

   if (strncmp (key, "batch_size", 20) == 0)
   {
      if (value <= 0)
         give_up (ERR_err, "Batch size must be positive\n");
      batch_size = value;
      return;
   }

   if (strncmp (key, "fou_sum_MC", 20) == 0)
   {
      if (value != 0 && value != 1)
         give_up (ERR_err,
                  "Summation of FG allelle must be MC (1) or exact (0)\n");
      fou_sum_MC = value;
      return;
   }

   if (strncmp (key, "gxg_interaction", 20) == 0)
   {
      if (value != 0 && value != 1)
         give_up (ERR_err,
                  "GxG interaction must be present (1) or absent (0)\n");
      gxg_interaction = value;
      return;
   }

   if (strncmp (key, "n_fou_sum_MC", 20) == 0)
   {
      if (value <= 0)
         give_up (ERR_err, "MC Realizations of FG allele must be positive\n");
      n_fou_sum_MC = value;
      return;
   }

   if (strncmp (key, "ScorIt", 20) == 0)
   {
      if (value <= 0)
         give_up (ERR_err, "Spacing of MCMC iterations must be positive\n");
      ScorIt = value;
      return;
   }

   if (strncmp (key, "null_it", 20) == 0)
   {
      if (value <= 0)
         give_up (ERR_err, "number must be positive\n");
      null_it = value;
      return;
   }

   if (strncmp (key, "l_marker_qtl2", 20) == 0)
   {
      if (value < 0)
         give_up (ERR_err, "index of left marker must be non-negative\n");
      l_marker_qtl2 = value;
      return;
   }

   if (strncmp (key, "l_marker_qtl1", 20) == 0)
   {
      if (value < 0)
         give_up (ERR_err, "index of left marker must be non-negative\n");
      l_marker_qtl1 = value;
      return;
   }

   if (strncmp (key, "debug_level", 20) == 0)
   {
      if (value < 0)
         give_up (ERR_err, "Debug level must be non-negative\n");
      debug_level = value;
      return;
   }

   give_up (ERR_err, "set_twoqtl_parm_int: key not recognized\n");
}

int get_twoqtl_parm_int (char *key)
{
   if (strncmp (key, "n_qtl", 20) == 0)
   {
      if (n_qtl == 1 || n_qtl == 2)
         return n_qtl;
      else
         give_up (ERR_err, "n_qtl not set\n");
   }

   if (strncmp (key, "n_batch", 20) == 0)
   {
      if (n_batch <= 0)
      {
         fprintf (stdout, " n_batch was set 20 by default\n");
         set_twoqtl_parm_int ("n_batch", 20);
      }
      return n_batch;
   }

   if (strncmp (key, "batch_size", 20) == 0)
   {
      if (batch_size > 0)
         return batch_size;
      else
         give_up (ERR_err, "batch_size not set\n");
   }

   if (strncmp (key, "fou_sum_MC", 20) == 0)
   {
      if (fou_sum_MC == 0 || fou_sum_MC == 1)
         return fou_sum_MC;
      else
         give_up (ERR_err, "fou_sum_MC not set\n");
   }
   if (strncmp (key, "gxg_interaction", 20) == 0)
   {
      if (gxg_interaction != 0 && gxg_interaction != 1)
      {
         fprintf (stdout, " gxg interaction was set absent (0) by default\n");
         set_twoqtl_parm_int ("gxg_interaction", 0);
      }
      return gxg_interaction;
   }

   if (strncmp (key, "n_fou_sum_MC", 20) == 0)
   {
      if (n_fou_sum_MC > 0)
         return n_fou_sum_MC;
      else
         give_up (ERR_err, "n_fou_sum_MC not set\n");
   }

   if (strncmp (key, "l_marker_qtl1", 20) == 0)
   {
      if (l_marker_qtl1 >= 0)
         return l_marker_qtl1;
      else
         give_up (ERR_err, "l_marker_qtl1 not set\n");
   }

   if (strncmp (key, "l_marker_qtl2", 20) == 0)
   {
      if (l_marker_qtl2 >= 0)
         return l_marker_qtl2;
      else
         give_up (ERR_err, "l_marker_qtl2 not set\n");
   }

   if (strncmp (key, "ScorIt", 20) == 0)
      return ScorIt;

   if (strncmp (key, "null_it", 20) == 0)
   {
      if (null_it <= 0)
      {
         fprintf (stdout, " null_it was set 10 by default\n");
         set_twoqtl_parm_int ("null_it", 10);
      }
      return null_it;
   }

   if (strncmp (key, "debug_level", 20) == 0)
   {
      if (debug_level < 0)
      {
         fprintf (stdout, " debug_level was set 0 by default\n");
         set_twoqtl_parm_int ("debug_level", 0);
      }
      return debug_level;
   }

   give_up (ERR_err, "get_twoqtl_parm_int: key not recognized\n");
   return 0;
}

void free_twoqtl_parm (void)
{
   if (geno_means_1 != NULL)
      free (geno_means_1);
   if (geno_means_2 != NULL)
      free (geno_means_2);
   if (geno_means_AA != NULL)
      free (geno_means_AA);
   if (geno_means_Aa != NULL)
      free (geno_means_Aa);
   if (geno_means_aa != NULL)
      free (geno_means_aa);
   return;
}
