/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/*
 NAME
    update_S_trait, free_update_S_trait, alloc_S_trait, free_S_trait

 SYNOPSIS

    #include "twoqtl.h"
    void update_S_trait (int comp_no, int nsw, int debug_level, 
                         char ***S_trait_1, char ***S_trait_0);
    void free_update_S_trait (void);

    char ***alloc_S_trait (int n_pos, int *size_c);
    void free_S_trait (char ***S); 

 DESCRIPTION                                                               

    update_S_trait samples the segregation indicators at the trait 
    positions, conditional on the segregation indicators at the marker
    positions in COMPONENT comp_no.

    print S_trait and S_markers for debug_level 1         
    print them split into each marker interval for debug_level 2 and 3  
    print S_parm (parameters for sampling)  for debug_level 3

    alloc_S_trait allocates S_trait_1 and S_trait_0, which are input 
    parameters of update_S_trait, and free_S_trait free them.

    free_update_S_trait frees all the static data structures used here.
    
 INPUT

    comp_no      component number, 0, 1, ..., n_compnts - 1          
    nsw          number of meioses that need to be updated             
    debug_level  0, 1, 2, 3

    S_trait_1    paternal segregation indicators at trait positions      
    S_trait_0    maternal segregation indicators at trait positions      

 GLOBAL VARIABLES USED
    LocPos, meiosis, NLocPos, NMarks, NTPosInt, xx1, xx0 

 SEE ALSO
    update_S_in and update_S_out below for detailed algorithms

 HISTORY
    initial code done                      Feb  7, 2005 
    updated to handle componentwise        May 16, 2005      
    document revised                       Jul  1, 2005      

 AUTHOR
    Written by Yun Ju Sung
*/

#include "twoqtl.h"

/* parameters for S_trait in each marker interval
   note: for the left end, index and recomb are reversed in order
         (reflecting the order of sampling)                      */
struct S_parm
{
   int  *index;                         /* trait positions in S_trait              */
   int   end;                           /* 1 for outside, 0 for inside             */
   double *recomb;                      /* recombinations between trait positions  */
   double ****prob;                     /* P(S_i, S_(i+1) | S_left = 0 or 1)       */
};
typedef struct S_parm S_parm_t;

static void setup_S_parm (void);
static void setup_prob (double *****prob_p, int n_prob, double *recomb);

static void fill_S_markers (char **S_1, char **S_0, int comp_no);
static void update_S_out (char **S_1, char **S_0, int intv_no, int comp_no);
static void update_S_in (char **S_1, char **S_0, int intv_no, int comp_no);

/* printing functions for debugging */
static void print_S_parm (int i);
static void print_single_marker (int mark_no, int comp_no,
                                 char **S_1, char **S_0);
static void print_S_trait (char **S_1, char **S_0, int intv_no, int comp_no);

static S_parm_t *S_trait_parm = NULL;

static int *mbase = NULL;               /* first index of each comp in meiosis */
static int *n_meiosis;                  /* number of meioses in each component */
static int *ibase = NULL;
static int *size_c = NULL;
static int *mark_index = NULL;          /* index of markers in S_trait */

/* ............... main function ...................... */

void update_S_trait (int comp_no, int nsw, int debug_level,
                     char ***S_trait_1, char ***S_trait_0)
{

   int   i;                             /* index for interval number */
   char **S_1 = S_trait_1[comp_no];
   char **S_0 = S_trait_0[comp_no];

/*   debug_level = 3; */
   if (!S_trait_parm)
   {
      setup_S_parm ();

      mbase = get_mbase (nsw);
      n_meiosis = get_n_meiosis (nsw);

      ibase = get_ibase ();
      size_c = get_size_c ();
   }

   fill_S_markers (S_1, S_0, comp_no);

   for (i = 0; i <= NMarks; ++i)
   {
      if (i == 0 || i == NMarks)
         update_S_out (S_1, S_0, i, comp_no);
      else
         update_S_in (S_1, S_0, i, comp_no);

      if (debug_level > 1)
         print_S_trait (S_1, S_0, i, comp_no);

      if (debug_level == 3)
         print_S_parm (i);
   }

   if (debug_level > 0)
   {
      print_S_marker (comp_no);
      print_S (S_1, S_0, NLocPos, comp_no);
   }

   return;
}

void fill_S_markers (char **S_1, char **S_0, int comp_no)
{
   int  *S_marker_1, *S_marker_0, index, i, j, j_nghd;
   int   comp_size = size_c[comp_no];
   char *S_trait_1, *S_trait_0;

   for (i = 0; i < NMarks; ++i)
   {
      S_marker_1 = xx1[i + 1];
      S_marker_0 = xx0[i + 1];

      index = mark_index[i];
      S_trait_1 = S_1[index];
      S_trait_0 = S_0[index];

      for (j = 0; j < comp_size; ++j)
      {
         j_nghd = nghd_index (j, comp_no);
         S_trait_1[j] = (char) S_marker_1[j_nghd];
         S_trait_0[j] = (char) S_marker_0[j_nghd];
      }
   }
   return;
}

/******************************************************************************
 *  NAME                                                                      *
 *    update_S_out (char **S_1, char **S_0, int intv_no, int comp_no)         *
 *                                                                            *
 *  DESCRIPTION                                                               *
 *                                                                            *
 *     Sample S_1, ..., S_n (S's at the trait positions) from                 *
 *     prob(S_1, ..., S_n | S_marker)                                         *
 *                                                                            *
 *  ALGORITHM                                                                 *
 *                                                                            *
 *    sample away from the marker                                             *
 *                                                                            *
 *    left: sample S_n, then S_(n-1) ... S_1                                  *
 *                                                                            *
 *              S_1    S_2      ....               S_n    S_marker            *
 *             --+------+------+------+------+------+------+-----             *
 *                                                                            *
 *    right: sample S_1 then S_2     ... S_n                                  *
 *                                                                            *
 *            S_marker S_1    S_2  ..                     S_n                 *
 *             --+------+------+------+------+------+------+-----             *
 *                                                                            *
 *                                                                            *
 *****************************************************************************/
void update_S_out (char **S_1, char **S_0, int intv_no, int comp_no)
{

   int  *index = S_trait_parm[intv_no].index;
   int   m_start = mbase[comp_no];
   int   m_stop = m_start + n_meiosis[comp_no];
   int   mark_no = (intv_no == 0) ? 1 : intv_no;

   int   k, j, j_comp, S_before;
   int   m, i;

   double *recomb = S_trait_parm[intv_no].recomb;
   char **S;

   for (i = 0; i < NTPosInt[intv_no]; ++i)
      for (m = m_start; m < m_stop; ++m)
      {
         k = meiosis[m].sex;
         j = (meiosis[m].ind).index;
         j_comp = comp_index (j, comp_no);

         S = (k == 0) ? S_0 : S_1;

         if (i == 0)
            S_before = (k == 0) ? xx0[mark_no][j] : xx1[mark_no][j];
         else
            S_before = S[index[i - 1]][j_comp];

         S[index[i]][j_comp] = (rans () <= recomb[i]) ? (!S_before) : S_before;
      }

   return;
}

/******************************************************************************
 *  NAME                                                                      *
 *    update_S_in  (char **S_1, char **S_0, int intv_no, int comp_no)         *
 *                                                                            *
 *  DESCRIPTION                                                               *
 *                                                                            *
 *     Sample S_1, ..., S_n (S's at the trait positions) from                 *
 *     prob(S_1, ..., S_n | S_left, S_right)                                  *
 *                                                                            *
 *      S_left  S_1    S_2      ....               S_n    S_right             *
 *      --+------+------+------+------+------+------+------+-----             *
 *                                                                            *
 *  ALGORITHM                                                                 *
 *                                                                            *
 *    sample backward (first S_n, next S_(n-1), ...., then S_1)               *
 *    each S_i from P(S_i | S_left, S_(i+1))                                  *
 *                                                                            *
 *    P(S_1, ..., S_n | S_left, S_right)                                      *
 *    = P(S_n | S_left, S_right) P(S_(n-1)| S_left, S_n) ...P(S_1|S_left, S_2)*
 *                                                                            *
 *    P(S_i | S_left, S_(i+1)) is proportional to  P(S_i, S_(i+1) | S_left)   *
 *    where the latter is prob[S_left][i][S_i][S_(i+1)]                       *
 *                                                                            *
 *****************************************************************************/
void update_S_in (char **S_1, char **S_0, int intv_no, int comp_no)
{
   int  *index = S_trait_parm[intv_no].index;
   int   m_start = mbase[comp_no];
   int   m_stop = m_start + n_meiosis[comp_no];

   int   k, j, j_comp, S_left, S_right;
   int   m, i, locus;

   double cond_prob_0, cond_prob_1;
   double ***prob;
   char **S;

   for (m = m_start; m < m_stop; ++m)
   {
      k = meiosis[m].sex;
      j = (meiosis[m].ind).index;
      j_comp = comp_index (j, comp_no);

      S = (k == 0) ? S_0 : S_1;
      S_left = (k == 0) ? xx0[intv_no][j] : xx1[intv_no][j];
      S_right = (k == 0) ? xx0[intv_no + 1][j] : xx1[intv_no + 1][j];

      prob = S_trait_parm[intv_no].prob[S_left];

      for (i = NTPosInt[intv_no] - 1; i > -1; --i)
      {
         locus = index[i];              /* trait locus position */

         if (i != NTPosInt[intv_no] - 1)
            S_right = S[locus + 1][j_comp];

         cond_prob_0 = prob[i][0][S_right];
         cond_prob_1 = prob[i][1][S_right];

         cond_prob_0 /= cond_prob_0 + cond_prob_1; /* normalize */

         S[locus][j_comp] = (rans () <= cond_prob_0) ? 0 : 1;
      }
   }

   return;
}

/* ............... setup functions .................... */

/*   setup S_trait_parm                       ***********/
/*   only called once by update_S_trait ()    ***********/
/******************************************************************************
 *                                                                            *
 *     S_trait_parm[i]                                                        *
 *                                                                            *
 *        i = 0           for left end of the marker map                      *
 *        i = NMarks      for right end of the marker map                     *
 *        0 < i < NMarks  for intervals inside the marker map                 *
 *                                                                            *
 *     for each i interval                                                    *
 *                                                                            *
 *        number of trait positions is NTPosInt                               *
 *                                                                            *
 *        recomb are recombination fractions between trait positions          *
 *           with length NTPosInt (for i = 0 and NMarks) or NTPosInt + 1      *
 *           (for 0 < i < NMarks)                                             *
 *                                                                            *
 *                 S_M(1)        S_M(2)                    S_M(NMarks)        *
 *    ----+--+--+---|---+--+--+---|---+--+--+---|---+--+--+---|---+--+--+---  *
 *        0  1  2   3   4  5  6   7   8  9  10  :index in S_trait             *
 *      NTPosInt[0]   NTPosInt[1]                                             *
 *                                                                            *
 *****************************************************************************/
void setup_S_parm (void)
{
   int   i, j, m;
   int   n_recomb;
   int   baz;

   int   k = 0;
   int   l = 0;
   double bar;

   S_trait_parm = malloc ((NMarks + 1) * sizeof (S_parm_t));
   alloc_chk (WHERE, S_trait_parm);

   mark_index = malloc (NMarks * sizeof (int));
   alloc_chk (WHERE, mark_index);

   for (i = 0; i <= NMarks; ++i)
   {
      S_parm_t *foo = S_trait_parm + i;

      /**** define  index    ****/
      foo->index = (int *) malloc (NTPosInt[i] * sizeof (int));
      alloc_chk (WHERE, foo->index);

      for (j = 0; j < NTPosInt[i]; ++j)
         foo->index[j] = l++;

      /**** define  end      ****/
      if (i == 0 || i == NMarks)
      {
         foo->end = 1;
         n_recomb = NTPosInt[i];
      }
      else
      {
         foo->end = 0;
         n_recomb = NTPosInt[i] + 1;
      }

      /**** define  recomb   ****/
      foo->recomb = (double *) malloc (n_recomb * sizeof (double));
      alloc_chk (WHERE, foo->recomb);

      for (j = 0; j < n_recomb; ++j)
      {
         foo->recomb[j] = dist2rf (LocPos[k + 1].pos[MAL] - LocPos[k].pos[MAL]);
         ++k;
      }

      /* the orders of recomb and index are reversed for the left end       */
      /* reflecting the order of sampling S_trait, in update_S_trait_out () */
      if (i == 0)
         for (j = 0, m = NTPosInt[0] - 1; j < m; ++j, --m)
         {
            baz = foo->index[j];
            foo->index[j] = foo->index[m];
            foo->index[m] = baz;

            bar = foo->recomb[j];
            foo->recomb[j] = foo->recomb[m];
            foo->recomb[m] = bar;
         }

      /**** define prob  ****/
      if (i == 0 || i == NMarks)
         foo->prob = NULL;
      else
         setup_prob (&(foo->prob), NTPosInt[i], foo->recomb);

      if (i < NMarks)
         mark_index[i] = l++;
   }

   return;
}

/*  setup prob in S_trait_parm               */
/*  only called once by setup_S_parm ()      */

/******************************************************************************
 *  NAME                                                                      *
 *    setup_prob (double *****prob_p, int n_prob, double *recomb)             *
 *                                                                            *
 *  OUTPUT                                                                    *
 *      (*prob_p)[2][n_prob][2][2]                                            *
 *                                                                            *
 *  INPUT                                                                     *
 *      recomb[n_prob+1] (recombination fractions between S_i and S_(i+1))    *
 *      n_prob           (number of trait positions in an interval)           *
 *                                                                            *
 *  DESCRIPTION                                                               *
 *                                                                            *
 *     Let S_left and S_right be segregation indicators at the left marker    *
 *     and the right marker and S_1, ..., S_n be segregation indicators at    *
 *     trait positions between the two.                                       *
 *                                                                            *
 *      S_left  S_1    S_2      ....               S_n    S_right             *
 *      --+------+------+------+------+------+------+------+-----             *
 *                                                                            *
 *     Then                                                                   *
 *                                                                            *
 *        prob[S_left][i][S_i][S_(i+1)] = Prob (S_i, S_(i+1) | S_left)        *
 *                                                                            *
 *     where i is 1,...,n (with S_(n+1) = S_right).                           *
 *                                                                            *
 *  ALGORITHM                                                                 *
 *                                                                            *
 *     Prob (S_i, S_(i+1) | S_left) = Prob (S_i | S_left) Prob(S_(i+1) | S_i) *
 *                                                                            *
 *     where                                                                  *
 *                                                                            *
 *        Prob (S_i | S_left)                                                 *
 *                                                                            *
 *          i = 1         |S_1 - S_left| is Bernoulli (recomb[0])             *
 *          i > 1         sum_(i-1) Prob(S_(i-1), S_i | S_left)               *
 *                                                                            *
 *        Prob (S_(i+1) | S_i)                                                *
 *                                                                            *
 *                    |S_(i+1) - S_i| is Bernoulli (recomb[i])                *
 *                                                                            *
 *****************************************************************************/

void setup_prob (double *****prob_p, int n_prob, double *recomb)
{
   int   i;                             /* index of locus           */
   int   k;                             /* index of S_left          */
   int   l;                             /* index of S_i             */
   int   m;                             /* index of S_(i+1)         */

   /* ---------------------------------------------------- */
   /*   allocate                                           */
   /* ---------------------------------------------------- */

   (*prob_p) = (double ****) malloc (2 * sizeof (double ***));
   alloc_chk (WHERE, (*prob_p));

   for (k = 0; k < 2; ++k)
      (*prob_p)[k] = alloc_dbl3d_two (n_prob, 2, 2);

   /* ---------------------------------------------------- */
   /*   define                                             */
   /* ---------------------------------------------------- */

   /** P(S_i, S_(i+1) | S_left) = P(S_i | S_left) P(S_(i+1) | S_i) **/

   for (i = 0; i < n_prob; ++i)
   {
      /****** prob[0]:    S_left = 0   ******/

      /**** (1) compute P(S_i | S_left)   ****/
      if (i == 0)
      {
         for (m = 0; m < 2; ++m)
         {
            (*prob_p)[0][i][1][m] = recomb[0];
            (*prob_p)[0][i][0][m] = 1 - recomb[0];
         }
      }
      else
      {
         for (m = 0; m < 2; ++m)
         {
            (*prob_p)[0][i][m][0] =
               (*prob_p)[0][i - 1][0][m] + (*prob_p)[0][i - 1][1][m];
            (*prob_p)[0][i][m][1] = (*prob_p)[0][i][m][0];
         }
      }

      /**** (2) multiply P(S_(i+1) | S_i)   ****/
      for (l = 0; l < 2; ++l)
         for (m = 0; m < 2; ++m)
            if (l != m)
               (*prob_p)[0][i][l][m] *= recomb[i + 1];
            else
               (*prob_p)[0][i][l][m] *= 1 - recomb[i + 1];

      /****** prob[1]:    S_left = 1   ******/
      for (l = 0; l < 2; ++l)
         for (m = 0; m < 2; ++m)
            (*prob_p)[1][i][l][m] = (*prob_p)[0][i][!l][!m];
   }

   return;
}

/* .......... printing functions ...................... */

void print_S_parm (int i)
{
   int   k, j, l;
   int   n_recomb;

   S_parm_t foo = S_trait_parm[i];

   fprintf (stderr, "\n *** S_trait_parm[%d] ****\n", i);

   /**** print  index    ****/
   fprintf (stderr, "\n index: ");
   for (j = 0; j < NTPosInt[i]; ++j)
      fprintf (stderr, "%3d", foo.index[j]);

   /**** print  end      ****/
   fprintf (stderr, "\n end:%3d", foo.end);

   /**** print  recomb   ****/
   if (foo.end == 1)
      n_recomb = NTPosInt[i];
   else
      n_recomb = NTPosInt[i] + 1;

   fprintf (stderr, "\n recomb: ");
   for (j = 0; j < n_recomb; ++j)
      fprintf (stderr, "%10.3f", foo.recomb[j]);

   /**** print prob  ****/
   if (foo.prob == NULL)
      fprintf (stderr, "\n prob is assigned as NULL\n");
   else
   {
      for (k = 0; k < 2; ++k)
      {
         fprintf (stderr, "\n prob[%d]:\n", k);
         for (l = 0; l < 2; ++l)
         {
            for (j = 0; j < NTPosInt[i]; ++j)
               fprintf (stderr, "%7.3f%6.3f", foo.prob[k][j][l][0],
                        foo.prob[k][j][l][1]);
            fprintf (stderr, "\n");
         }
         fprintf (stderr, "sum to:\n");
         for (j = 0; j < NTPosInt[i]; ++j)
            fprintf (stderr, " %e", foo.prob[k][j][0][0] +
                     foo.prob[k][j][0][1] +
                     foo.prob[k][j][1][0] + foo.prob[k][j][1][1]);
         fprintf (stderr, "\n");
      }
   }

   return;
}

void print_single_marker (int mark_no, int comp_no, char **S_1, char **S_0)
{
   int  *S_marker_1 = xx1[mark_no];
   int  *S_marker_0 = xx0[mark_no];

   int   index = mark_index[mark_no - 1];
   char *S_trait_1 = S_1[index];
   char *S_trait_0 = S_0[index];

   int   comp_size = size_c[comp_no];
   int   j_start = ibase[comp_no];
   int   j_stop = j_start + comp_size;
   int   j;

   fprintf (stderr, " m%d", mark_no);
   for (j = j_start; j < j_stop; ++j)
      fprintf (stderr, "%3d%2d", S_marker_1[j], S_marker_0[j]);
   fprintf (stderr, "\n");

   fprintf (stderr, " %2d", index);
   for (j = 0; j < comp_size; ++j)
      fprintf (stderr, "%3d%2d", (int) S_trait_1[j], (int) S_trait_0[j]);
   fprintf (stderr, "\n");

   return;
}

void print_S_trait (char **S_1, char **S_0, int intv_no, int comp_no)
{
   int  *index, n_index, mark_no;
   int   j_stop, locus, i, j;
   char *S_1_locus, *S_0_locus;

   index = S_trait_parm[intv_no].index;
   n_index = NTPosInt[intv_no];

   fprintf (stderr, "=== print_S_trait for interval %d and component %d ===\n",
            intv_no, comp_no);

   mark_no = (intv_no == 0) ? 1 : intv_no;
   print_single_marker (mark_no, comp_no, S_1, S_0);

   j_stop = size_c[comp_no];

   for (i = 0; i < n_index; ++i)
   {
      locus = index[i];
      fprintf (stderr, " %2d", locus);

      S_1_locus = S_1[locus];
      S_0_locus = S_0[locus];
      for (j = 0; j < j_stop; ++j)
         fprintf (stderr, "%3d%2d", (int) S_1_locus[j], (int) S_0_locus[j]);
      fprintf (stderr, "\n");
   }

   if (S_trait_parm[intv_no].end == 0)
      print_single_marker (mark_no + 1, comp_no, S_1, S_0);

   return;
}

char ***alloc_S_trait (int n_pos, int *n_people)
{
   int   comp_size, i;
   char ***S;

   S = (char ***) malloc (n_compnts * sizeof (char **));
   alloc_chk (WHERE, S);

   for (i = 0; i < n_compnts; i++)
   {
      comp_size = n_people[i];
      S[i] = alloc_char2d (n_pos, comp_size);
      memset (S[i][0], -1, n_pos * comp_size * sizeof (char)); /*init. to -1 */
   }

   return S;
}

void free_S_trait (char ***S)
{
   int   i;

   for (i = 0; i < n_compnts; i++)
      free_char2d (S[i]);
   free (S);
   return;
}

void free_update_S_trait (void)
{
   int   i, k;
   double ****prob;

   for (i = 0; i <= NMarks; ++i)
   {
      S_parm_t S_parm = S_trait_parm[i];
      free (S_parm.index);
      free (S_parm.recomb);
      if (S_parm.prob != NULL)
      {
         prob = S_parm.prob;
         for (k = 0; k < 2; ++k)
            free_dbl3d (prob[k]);
      }
   }
   free (S_trait_parm);
   free (mark_index);

   return;
}
